"""
Exposes a Thing with a property that returns the current 
UNIX timestamp and a couple of events that fire periodically 
with the current date and UNIX timestamp.
"""

import asyncio
import datetime
import functools
import json
import logging
import pprint
import time

from wotpy.wot.enums import DataType
from wotpy.wot.td import ThingDescription

_DESCRIPTION = {
    "id": "urn:org:fundacionctic:thing:clock",
    "name": "Clock Thing",
    "properties": {
        "time": {
            "type": DataType.INTEGER,
            "readOnly": True
        }
    },
    "events": {
        "timeEvent": {
            "data": {
                "type": DataType.INTEGER
            }
        },
        "dateEvent": {
            "data": {
                "type": DataType.STRING
            }
        }
    }
}

_TIMEOUT = 10

_logger = logging.getLogger(__name__)


def _time_millis():
    return int(time.time() * 1e3)


async def _time_handler():
    return _time_millis()


def _emit_time(exposed_thing):
    payload = _time_millis()
    _logger.debug("Emitting timeEvent: %s", payload)
    exposed_thing.emit_event("timeEvent", payload)


def _emit_date(exposed_thing):
    payload = datetime.datetime.utcnow().isoformat()
    _logger.debug("Emitting dateEvent: %s", payload)
    exposed_thing.emit_event("dateEvent", payload)


async def _emitter(emit_func, interval):
    interval = float(interval)

    try:
        while True:
            next_time = time.time() + interval
            emit_func()
            sleep = next_time - time.time()
            await asyncio.sleep(max(0, sleep))
    except asyncio.CancelledError:
        _logger.debug("One last event before cancelling: %s", emit_func)
        emit_func()


async def app(wot, conf, loop, interval=2):
    _logger.info("Starting clock app with interval: %s seconds", interval)

    _logger.info(
        "Producing Thing:\n%s",
        pprint.pformat(_DESCRIPTION))

    exposed_thing = wot.produce(json.dumps(_DESCRIPTION))
    exposed_thing.set_property_read_handler("time", _time_handler)
    exposed_thing.expose()

    _logger.debug(
        "Exposed Thing:\n%s",
        pprint.pformat(ThingDescription.from_thing(exposed_thing.thing).to_dict()))

    emit_time = functools.partial(_emit_time, exposed_thing)
    emit_date = functools.partial(_emit_date, exposed_thing)

    time_task = asyncio.ensure_future(_emitter(emit_time, interval=interval))
    date_task = asyncio.ensure_future(_emitter(emit_date, interval=interval))

    try:
        await asyncio.gather(time_task, date_task)
    except asyncio.CancelledError:
        await asyncio.wait_for(asyncio.gather(time_task, date_task), _TIMEOUT)



"""
Video camera aggregator with simple face detection capabilities.
"""

import asyncio
import base64
import functools
import json
import logging
import pprint
import sys
import time

import cv2
import face_recognition
import numpy as np
from wotemu.monitor.utils import write_metric
from wotemu.utils import consume_from_catalogue, wait_node
from wotpy.wot.td import ThingDescription

_QUEUE_MAXSIZE = 5
_QUEUE_MAX_SECONDS = 10
_DEFAULT_CAMERA_ID = "urn:org:fundacionctic:thing:wotemu:camera"
_DEFAULT_FRAME_EVENT = "jpgVideoFrame"
_PTZ_ACTION = "controlPTZ"
_METRIC_LATENCY = "detection_latency"

_DESCRIPTION = {
    "id": "urn:org:fundacionctic:thing:wotemu:detector",
    "name": "Camera detection",
    "properties": {
        "cameras": {
            "readOnly": True,
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "servient_host": {
                        "type": "string"
                    },
                    "thing_id": {
                        "type": "string"
                    },
                    "frame_event": {
                        "type": "string"
                    }
                },
                "required": ["servient_host"]
            }
        },
        "latestDetections": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "cameraId": {
                        "type": "string"
                    },
                    "jpgB64": {
                        "type": "string"
                    },
                    "faceLocations": {
                        "type": "array"
                    },
                    "unixTime": {
                        "type": "number"
                    }
                },
                "required": ["cameraId", "jpgB64", "faceLocations", "unixTime"]
            }
        }
    }
}

_logger = logging.getLogger(__name__)


def _detect(queue_item):
    b64_jpg = queue_item["b64_jpg"]
    jpg_bytes = base64.b64decode(b64_jpg)
    frame_arr = cv2.imdecode(np.frombuffer(jpg_bytes, np.uint8), -1)
    return face_recognition.face_locations(frame_arr)


async def _control_camera_ptz(wot, conf, camera_id, lock, **kwargs):
    try:
        await lock.acquire()

        _logger.debug("Invoking PTZ control: %s", camera_id)

        splitted = camera_id.split("::")
        servient_host = splitted[0].strip()
        thing_id = splitted[1].strip()

        camera_thing = await consume_from_catalogue(
            wot=wot,
            port_catalogue=conf.port_catalogue,
            servient_host=servient_host,
            thing_id=thing_id)

        await camera_thing.actions[_PTZ_ACTION].invoke(kwargs)

        _logger.debug("Finished PTZ control invocation: %s", camera_id)
    except Exception as ex:
        _logger.warning("Failed updating PTZ controls: %s", repr(ex))
    finally:
        lock.release()


async def _process_queue(wot, conf, detection_queue, store, timeout_get=3.0):
    locks = {}

    while True:
        try:
            queue_item = await asyncio.wait_for(detection_queue.get(), timeout_get)
        except asyncio.TimeoutError:
            _logger.debug("Timeout waiting for queue item")
            continue

        b64_jpg = queue_item["b64_jpg"]
        cam_id = queue_item["cam_id"]
        time_arrival = queue_item["time_arrival"]
        time_capture = queue_item["time_capture"]

        if (time.time() - time_arrival) >= _QUEUE_MAX_SECONDS:
            _logger.info("Dropping old item (>%s s.)", _QUEUE_MAX_SECONDS)
            continue

        _logger.debug(
            "[%s] Received B64-encoded JPG (%s) (%s KB)",
            cam_id,
            b64_jpg.__class__.__name__,
            round(sys.getsizeof(b64_jpg) / 1024.0, 1))

        time_pre_detect = time.time()

        loop = asyncio.get_running_loop()
        detect = functools.partial(_detect, queue_item=queue_item)
        face_locations = await loop.run_in_executor(None, detect)
        no_faces = not face_locations or len(face_locations) == 0

        time_detect = time.time()

        _logger.log(
            logging.DEBUG if no_faces else logging.INFO,
            "[%s] Face locations: %s",
            cam_id, face_locations)

        await write_metric(key=_METRIC_LATENCY, data={
            "time_capture": time_capture,
            "time_arrival": time_arrival,
            "time_detect": time_detect,
            "latency_frame": time_arrival - time_capture,
            "latency_detect": time_detect - time_arrival,
            "detect_cost": time_detect - time_pre_detect,
            "camera_id": cam_id
        })

        if no_faces:
            continue

        store_idx = next((
            idx for idx, item in enumerate(store)
            if item["cameraId"] == cam_id), None)

        if store_idx is not None:
            store.pop(store_idx)

        store.append({
            "cameraId": cam_id,
            "jpgB64": b64_jpg,
            "faceLocations": face_locations,
            "unixTime": time_capture
        })

        if not locks.get(cam_id):
            _logger.debug("Initializing lock for %s", cam_id)
            locks[cam_id] = asyncio.Lock()

        lock = locks[cam_id]

        if lock.locked():
            _logger.debug("Skipping PTZ control for %s", cam_id)
            continue

        asyncio.ensure_future(_control_camera_ptz(
            wot=wot,
            conf=conf,
            camera_id=cam_id,
            lock=lock,
            time_capture=time_capture,
            time_arrival=time_arrival,
            time_detect=time_detect))


def _on_next(item, cam_id, detection_queue):
    try:
        _logger.debug("Putting new item in detection queue")

        detection_queue.put_nowait({
            "b64_jpg": item.data["b64_jpg"],
            "time_capture": item.data["time_capture"],
            "time_arrival": time.time(),
            "cam_id": cam_id
        })
    except asyncio.QueueFull:
        _logger.info("Detection queue is full")


def _on_error(err, cam_id, error_event):
    _logger.warning("[%s] Subscription error: %s", cam_id, err)
    error_event.set()


async def _subscribe_camera(wot, conf, camera, error_event, detection_queue):
    servient_host = camera["servient_host"]
    thing_id = camera.get("thing_id", _DEFAULT_CAMERA_ID)
    frame_event = camera.get("frame_event", _DEFAULT_FRAME_EVENT)
    cam_id = f"{servient_host} :: {thing_id}"

    await wait_node(conf=conf, name=servient_host)

    camera_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=conf.port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    on_next = functools.partial(
        _on_next,
        cam_id=cam_id,
        detection_queue=detection_queue)

    on_error = functools.partial(
        _on_error,
        cam_id=cam_id,
        error_event=error_event)

    return camera_thing.events[frame_event].subscribe(
        on_next=on_next,
        on_error=on_error)


async def _cancel_subs(subs, cancel_sleep=3):
    _logger.info("Cancelling subscriptions: %s", subs)

    for sub in subs:
        try:
            _logger.debug("Disposing: {}".format(sub))
            sub.dispose()
        except Exception as ex:
            _logger.debug("Error disposing of %s: %s", sub, ex)

    _logger.info("Waiting %s s. for subscriptions", cancel_sleep)
    await asyncio.sleep(cancel_sleep)


async def _subscribe_task(wot, conf, cameras, detection_queue):
    error_event = asyncio.Event()

    while True:
        _logger.info("Clearing error event")
        error_event.clear()

        try:
            _logger.info(
                "Subscribing to cameras:\n%s",
                pprint.pformat(cameras))

            subs = await asyncio.gather(*[
                _subscribe_camera(
                    wot=wot,
                    conf=conf,
                    camera=camera,
                    error_event=error_event,
                    detection_queue=detection_queue)
                for camera in cameras
            ], return_exceptions=False)

            await error_event.wait()
            _logger.warning("Detected error event")
        finally:
            try:
                _cancel_subs(subs)
            except:
                _logger.warning("Error cancelling subs", exc_info=True)


async def _identity(val):
    return val


async def app(wot, conf, loop, cameras):
    store = list()
    cameras = json.loads(cameras)
    detection_queue = asyncio.LifoQueue(maxsize=_QUEUE_MAXSIZE)

    _logger.info(
        "Producing Thing:\n%s",
        pprint.pformat(_DESCRIPTION))

    exposed_thing = wot.produce(json.dumps(_DESCRIPTION))

    exposed_thing.set_property_read_handler(
        "latestDetections",
        functools.partial(_identity, val=store))

    exposed_thing.set_property_read_handler(
        "cameras",
        functools.partial(_identity, val=cameras))

    exposed_thing.expose()

    _logger.debug(
        "Exposed Thing:\n%s",
        pprint.pformat(ThingDescription.from_thing(exposed_thing.thing).to_dict()))

    await asyncio.gather(
        _process_queue(
            wot=wot,
            conf=conf,
            detection_queue=detection_queue,
            store=store),
        _subscribe_task(
            wot=wot,
            conf=conf,
            cameras=cameras,
            detection_queue=detection_queue),
        return_exceptions=False)



import asyncio
import json
import logging
import os
import time
from collections import namedtuple

import aioredis
import docker
import pytest
from wotemu.report.reader import ReportDataRedisReader

TD_EXAMPLE = {
    "id": "urn:org:fundacionctic:thing:testthing",
    "name": "Test Thing",
    "properties": {
        "testProp": {
            "description": "Test property",
            "type": "number",
            "readOnly": False,
            "observable": True
        }
    }
}

_ENV_REDIS_TEST = "TEST_REDIS_URL"
_DOCKER_REDIS_IMAGE = "redis:5"
_DOCKER_REDIS_PORT = "6379/tcp"
_REDIS_DATA_FILE = "./redis_data.json"

RedisService = namedtuple(
    "RedisService",
    ["url", "container"])

_logger = logging.getLogger(__name__)


def start_redis_container():
    client = docker.from_env()
    assert client.ping(), "Error on Docker ping"

    _logger.debug("Pulling image: %s", _DOCKER_REDIS_IMAGE)

    client.images.pull(_DOCKER_REDIS_IMAGE)

    container = client.containers.run(
        image=_DOCKER_REDIS_IMAGE,
        detach=True,
        ports={_DOCKER_REDIS_PORT: None})

    _logger.debug("Created Redis container: %s", container.name)

    inspect_data = client.api.inspect_container(container.id)
    port = inspect_data["NetworkSettings"]["Ports"][_DOCKER_REDIS_PORT][0]["HostPort"]
    url = "redis://localhost:{}".format(port)

    _logger.debug("Redis container URL: %s", url)

    return RedisService(container=container, url=url)


def get_redis_service():
    redis_url = os.getenv(_ENV_REDIS_TEST, None)

    if redis_url:
        _logger.debug("Using Redis from environment: %s", redis_url)
        return RedisService(url=redis_url, container=None)

    _logger.warning(
        "Undefined Redis URL ($%s): Creating volatile Docker container",
        _ENV_REDIS_TEST)

    return start_redis_container()


async def wait_redis(redis_service, timeout=10.0):
    ini = time.time()

    while True:
        try:
            _logger.debug("Waiting for Redis service")
            redis = await aioredis.create_redis_pool(redis_service.url)
            assert await redis.ping()
            _logger.debug("Redis service ping OK")
            break
        except Exception as ex:
            _logger.debug("Error on Redis ping: %s", ex)

            if (time.time() - ini) > timeout:
                msg = "Redis service timeout ({} s)".format(timeout)
                _logger.warning(msg)
                raise Exception(msg)
        finally:
            try:
                redis.close()
                await redis.wait_closed()
            except:
                pass


@pytest.fixture
async def redis():
    try:
        redis_service = get_redis_service()
        await wait_redis(redis_service)
        redis = await aioredis.create_redis_pool(redis_service.url)
        await redis.flushdb()
        yield redis
        await redis.flushdb()
        redis.close()
        await redis.wait_closed()
    except Exception as ex:
        msg = "Error initiliazing Redis fixture: {}".format(ex)
        _logger.warning(msg)
        pytest.skip(msg)
    finally:
        try:
            redis_service.container.remove(force=True, v=True)
        except:
            pass


def _insert_zset(tr, item):
    assert item["type"] == "zset"

    for value_item in item["value"]:
        tr.zadd(
            key=item["key"],
            score=value_item[1],
            member=value_item[0])


@pytest.fixture
async def redis_loaded(redis):
    current_dir = os.path.dirname(os.path.realpath(__file__))
    data_file = os.path.join(current_dir, _REDIS_DATA_FILE)

    with open(data_file, "r") as fh:
        data = json.loads(fh.read())

    tr = redis.multi_exec()

    for item in data:
        _logger.debug(
            "Loading '%s' (type: %s) (size: %s)",
            item.get("key"), item.get("type"), len(item.get("value", [])))

        if item["type"] == "zset":
            _insert_zset(tr, item)

    await tr.execute()

    return redis


@pytest.fixture
async def redis_reader(redis_loaded):
    host, port = redis_loaded.connection.address
    db = redis_loaded.connection.db
    redis_url = f"redis://{host}:{port}/{db}"
    reader = ReportDataRedisReader(redis_url=redis_url)
    await reader.connect()
    yield reader
    await reader.close()


@pytest.fixture
def redis_test_data():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    data_file = os.path.join(current_dir, _REDIS_DATA_FILE)

    with open(data_file, "r") as fh:
        data = json.loads(fh.read())
        return RedisTestData(data)


class RedisTestData:
    def __init__(self, data):
        self._data = data

    def get_num_tasks(self):
        return len(set(
            item["key"]
            for item in self._data
            if item["key"].startswith("wotemu:info:")))

    def _next_task_with_data(self, redis_key):
        prefix = "wotemu:{}:".format(redis_key)

        return next(
            item["key"].split(":")[-1]
            for item in self._data
            if item["key"].startswith(prefix))

    def get_task_with_system_data(self):
        return self._next_task_with_data(redis_key="system")

    def get_task_with_packet_data(self):
        return self._next_task_with_data(redis_key="packet")

    def get_task_with_thing_data(self):
        return self._next_task_with_data(redis_key="thing")



import asyncio
import logging
from datetime import datetime
from pathlib import Path

from wotemu.cli.utils import find_stack_redis_port
from wotemu.report.builder import ReportBuilder
from wotemu.report.reader import ReportDataRedisReader

_logger = logging.getLogger(__name__)


async def _connect_and_build(redis_url, base_path, as_json, file_name):
    reader = ReportDataRedisReader(redis_url=redis_url)

    try:
        await reader.connect()
        builder = ReportBuilder(reader=reader)
        Path(base_path).mkdir(parents=True, exist_ok=True)

        if as_json:
            _logger.info("Writing report as raw JSON")

            await builder.write_report_dataset(
                base_path=base_path,
                file_name=file_name)
        else:
            _logger.info("Writing report as HTML page")
            await builder.write_report(base_path=base_path)
    finally:
        try:
            await reader.close()
        except:
            pass


def build_report(conf, out, stack, redis_url, as_json):
    if not stack and not redis_url:
        raise ValueError((
            "You must provide either an explicit Redis URL "
            "or the name of the stack. This command must "
            "be executed in a manager node if you choose "
            "to provide the name of the stack."
        ))

    if not redis_url:
        redis_port = find_stack_redis_port(stack=stack)
        redis_url = f"redis://127.0.0.1:{redis_port}"

    _logger.info("Using Redis URL: %s", redis_url)

    dtime = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
    file_name = f"wotemu_{dtime}"
    file_name = f"{file_name}_{stack}" if stack else file_name

    loop = asyncio.get_event_loop()

    loop.run_until_complete(_connect_and_build(
        redis_url=redis_url,
        base_path=out,
        as_json=as_json,
        file_name=file_name))



import lxml.etree
from wotemu.report.components.base import BaseComponent
from wotemu.report.components.figure_block import FigureBlockComponent


class ContainerComponent(BaseComponent):
    def __init__(self, elements, margin="mt-2"):
        self.elements = elements
        self.margin = margin

    def to_element(self):
        elements = []

        for item in self.elements:
            try:
                elements.append(item.to_element())
            except AttributeError:
                elements.append(item)

        container = lxml.etree.Element("div")

        for idx, el in enumerate(elements):
            row_class = f"row {self.margin}" if idx > 0 else "row"
            row = lxml.etree.Element("div", attrib={"class": row_class})
            col = lxml.etree.Element("div", attrib={"class": "col"})
            col.append(el)
            row.append(col)
            container.append(row)

        return container



"""
Mock video camera with naive motion detection capabilities.
"""

import asyncio
import base64
import collections
import importlib.resources
import json
import logging
import pprint
import random
import sys
import time

import cv2
import numpy as np
from wotemu.monitor.utils import write_metric
from wotpy.wot.td import ThingDescription

_VIDEO_PKG = "wotemu.apps.data"
_VIDEO_RESOURCE = "camera.mp4"
_MOTION_THRESHOLD = 25.0
_JPEG_QUALITY = 60
_PTZ_MU = 0.0
_PTZ_SIGMA = 1.0
_PTZ_LOOP_SLEEP = 0.1
_METRIC_PTZ_LATENCY = "ptz_latency"


_DESCRIPTION = {
    "id": "urn:org:fundacionctic:thing:wotemu:camera",
    "name": "Mock video camera",
    "actions": {
        "controlPTZ": {
            "safe": True,
            "idempotent": False,
            "input": {
                "type": "object"
            }
        }
    },
    "events": {
        "jpgVideoFrame": {
            "data": {
                "type": "object"
            }
        }
    }
}

_logger = logging.getLogger(__name__)


def _get_video_path():
    with importlib.resources.path(_VIDEO_PKG, _VIDEO_RESOURCE) as pth:
        return str(pth)


def _frames_generator():
    video_path = _get_video_path()

    try:
        _logger.info("Initializing cv2.VideoCapture: %s", video_path)
        cap = cv2.VideoCapture(video_path)
        frame_counter = 0

        while cap.isOpened():
            ret, frame = cap.read()
            frame_counter += 1

            if not ret:
                raise RuntimeError("Error reading video frame")

            if frame_counter == cap.get(cv2.CAP_PROP_FRAME_COUNT):
                _logger.debug("Resetting cv2.CAP_PROP_POS_FRAMES")
                frame_counter = 0
                cap.set(cv2.CAP_PROP_POS_FRAMES, 0)

            yield frame
    finally:
        try:
            cap.release()
        except:
            _logger.warning("Error releasing capture", exc_info=True)


def _get_motion_score(frames):
    frames_grey = [cv2.cvtColor(item, cv2.COLOR_BGR2GRAY) for item in frames]
    frames_flat = [item.flatten() for item in frames_grey]
    diff_arr = np.diff(frames_flat, axis=0)
    means_arr = [abs(np.mean(item)) for item in diff_arr]
    return np.mean(means_arr)


async def _video_motion_generator(target_fps=12, buf_size=None, motion_threshold=None):
    buf_size = buf_size or int(target_fps)
    motion_threshold = motion_threshold or _MOTION_THRESHOLD
    frames_buf = collections.deque([], buf_size)
    frame_sleep = 1.0 / float(target_fps)

    for frame in _frames_generator():
        frames_buf.append(frame)
        motion_score = _get_motion_score(frames_buf)
        yield frame, motion_score > motion_threshold
        await asyncio.sleep(frame_sleep)


def _to_b64_jpg(frame):
    params = [cv2.IMWRITE_JPEG_QUALITY, _JPEG_QUALITY]
    retval, img_bytes = cv2.imencode(".jpg", frame, params)
    assert retval
    return base64.b64encode(img_bytes).decode()


async def _video_frame_emitter(exposed_thing):
    async for frame, motion in _video_motion_generator():
        if motion:
            frame_jpg = _to_b64_jpg(frame)
            frame_kb = round(sys.getsizeof(frame_jpg) / 1024.0, 1)
            _logger.debug("Emitting jpgVideoFrame (%s KB)", frame_kb)
            assert isinstance(frame_jpg, str)

            exposed_thing.emit_event("jpgVideoFrame", {
                "time_capture": time.time(),
                "b64_jpg": frame_jpg
            })


async def _control_ptz(params):
    params_input = params["input"]
    params_input = params_input or {}

    _logger.debug("Serving PTZ invocation:\n%s", pprint.pformat(params_input))

    if params_input.get("time_capture"):
        latency_ptz = time.time() - float(params_input["time_capture"])
        params_input.update({"latency_ptz": latency_ptz})
        await write_metric(key=_METRIC_PTZ_LATENCY, data=params_input)

    sleep_secs = abs(random.gauss(_PTZ_MU, _PTZ_SIGMA))
    sleep_end = time.time() + sleep_secs

    while time.time() < sleep_end:
        await asyncio.sleep(_PTZ_LOOP_SLEEP)

    return params_input


async def app(wot, conf, loop):
    _logger.info("Starting mock video camera app")

    _logger.info(
        "Producing Thing:\n%s",
        pprint.pformat(_DESCRIPTION))

    exposed_thing = wot.produce(json.dumps(_DESCRIPTION))
    exposed_thing.set_action_handler("controlPTZ", _control_ptz)
    exposed_thing.expose()

    _logger.debug(
        "Exposed Thing:\n%s",
        pprint.pformat(ThingDescription.from_thing(exposed_thing.thing).to_dict()))

    await _video_frame_emitter(exposed_thing=exposed_thing)



import logging
import os
import warnings

import inflection
import yaml
from wotemu.config import (DEFAULT_CONFIG_VARS, DEFAULT_HOST_DOCKER_PROXY,
                           DEFAULT_HOST_REDIS, ConfigVars)
from wotemu.enums import NETEM_CONDITIONS, BuiltinApps, NetworkConditions
from wotemu.topology.compose import (BASE_IMAGE, IMAGE_ENV_VAR,
                                     get_broker_definition,
                                     get_docker_proxy_definition,
                                     get_generic_service_definition,
                                     get_network_definition,
                                     get_network_gateway_definition,
                                     get_node_definition, get_redis_definition,
                                     get_topology_definition)

_DEFAULT_CATALOGUE = DEFAULT_CONFIG_VARS[ConfigVars.PORT_CATALOGUE]
_DEFAULT_HTTP = DEFAULT_CONFIG_VARS[ConfigVars.PORT_HTTP]
_DEFAULT_WS = DEFAULT_CONFIG_VARS[ConfigVars.PORT_WS]
_DEFAULT_MQTT = DEFAULT_CONFIG_VARS[ConfigVars.PORT_MQTT]
_DEFAULT_COAP = DEFAULT_CONFIG_VARS[ConfigVars.PORT_COAP]

_logger = logging.getLogger(__name__)


class NodeResources:
    """Represents the hardware constraints and resources of a node."""

    def __init__(
            self, cpu_limit=None, mem_limit=None,
            cpu_reservation=None, mem_reservation=None, target_cpu_speed=None):
        if cpu_limit and target_cpu_speed:
            raise ValueError("Use either CPU limit or target CPU speed")

        self._cpu_limit = cpu_limit
        self._mem_limit = mem_limit
        self._cpu_reservation = cpu_reservation
        self._mem_reservation = mem_reservation
        self.target_cpu_speed = target_cpu_speed

    @property
    def cpu_limit(self):
        return str(self._cpu_limit) if self._cpu_limit else None

    @property
    def mem_limit(self):
        return str(self._mem_limit) if self._mem_limit else None

    @property
    def cpu_reservation(self):
        return str(self._cpu_reservation) if self._cpu_reservation else None

    @property
    def mem_reservation(self):
        return str(self._mem_reservation) if self._mem_reservation else None


class NodeApp:
    """Represents the arguments to run a WoT app in a node."""

    ARG_PATH = "--path"
    ARG_PARAM = "--func-param"
    ARG_HTTP = "--enable-http"
    ARG_WS = "--enable-ws"
    ARG_MQTT = "--enable-mqtt"
    ARG_COAP = "--enable-coap"

    def __init__(self, path, http=False, ws=False, mqtt=False, coap=False, params=None):
        self._path = path
        self.params = params if params else {}
        self._http = http
        self._ws = ws
        self._mqtt = mqtt
        self._coap = coap

    @property
    def path(self):
        bapp_members = [item for item in BuiltinApps]
        return self._path.value if self._path in bapp_members else self._path

    @property
    def app_args(self):
        parts = [self.ARG_PATH, self.path]

        for key, val in self.params.items():
            parts += [self.ARG_PARAM, str(key), str(val)]

        protocol_flags = [
            (self.enabled_http, self.ARG_HTTP),
            (self.enabled_ws, self.ARG_WS),
            (self.enabled_mqtt, self.ARG_MQTT),
            (self.enabled_coap, self.ARG_COAP)
        ]

        parts += [flag for enabled, flag in protocol_flags if enabled]

        return parts

    @property
    def enabled_http(self):
        return bool(self._http)

    @property
    def enabled_ws(self):
        return bool(self._ws)

    @property
    def enabled_mqtt(self):
        return bool(self._mqtt)

    @property
    def enabled_coap(self):
        return bool(self._coap)


class BaseNamedModel:
    """Base class of all models with a unique name."""

    def __init__(self, name):
        name_clean = inflection.underscore(name)

        if name != name_clean:
            _logger.warning("Converted name '%s' to '%s'", name, name_clean)

        self._name = name_clean

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    @property
    def name(self):
        return self._name


class Service(BaseNamedModel):
    """Represents any generic service (e.g. a database) that 
    may be required by a node of the emulation experiment."""

    def __init__(self, name, image, **kwargs):
        """Optional parameters are the same as client.services.create:
        https://docker-py.readthedocs.io/en/stable/services.html"""

        self.image = image
        self._params = kwargs
        super().__init__(name)

    @property
    def params(self):
        return self._params

    def get_node_network_name(self, node):
        return f"srvnet_{self.name}_{node.name}"

    def to_compose_dict(self, topology):
        return get_generic_service_definition(topology, self)


class Node(BaseNamedModel):
    """Represents an independent node that runs any given user-defined WoT application."""

    ENTRY_APP = "app"

    @staticmethod
    def _assert_broker(app, broker):
        if app.enabled_mqtt and not broker:
            raise ValueError("A broker should be defined when MQTT is enabled")

    @staticmethod
    def _assert_broker_network(broker, broker_network):
        if broker and broker_network and broker_network not in broker.networks:
            raise ValueError((
                "The node's broker ({}) is not "
                "attached to the node's broker network ({})"
            ).format(broker, broker_network))

    @staticmethod
    def _warn_broker_network_undefined(broker, broker_network):
        if broker and not broker_network and len(broker.networks) > 1:
            warnings.warn((
                "The node's broker ({}) is attached to "
                "multiple networks but the node's "
                "broker network has not been explicitly defined"
            ).format(broker), Warning)

    @staticmethod
    def _warn_broker_unrequired(broker, app):
        if broker and not app.enabled_mqtt:
            warnings.warn((
                "There is no need to set the node broker ({}) "
                "if MQTT is disabled in the node app ({})"
            ).format(broker, app), Warning)

    def __init__(
            self, name, app, networks, broker=None, broker_network=None,
            image=None, resources=None, scale=None, args_compose=None, services=None):
        self._assert_broker(app, broker)
        self._assert_broker_network(broker, broker_network)
        self._warn_broker_network_undefined(broker, broker_network)
        self._warn_broker_unrequired(broker, app)

        self.app = app
        self.networks = networks
        self.broker = broker
        self._broker_network = broker_network
        self._image = image
        self.resources = resources
        self.scale = scale
        self.args_compose = args_compose
        self._services = set(services) if services else set()
        super().__init__(name)

    def __repr__(self):
        return "{} (name={})".format(super().__repr__(), self.name)

    @property
    def image(self):
        return self._image or os.getenv(IMAGE_ENV_VAR, BASE_IMAGE)

    @property
    def broker_network(self):
        if not self.broker:
            return None

        if self._broker_network:
            assert self._broker_network in self.broker.networks
            return self._broker_network

        return self.broker.networks[0]

    @property
    def broker_host(self):
        if not self.broker:
            return None

        return "{}.{}".format(self.broker.name, self.broker_network.name)

    @property
    def cmd_app(self):
        return [self.ENTRY_APP] + list(self.app.app_args)

    @property
    def services(self):
        return self._services

    def link_service(self, srv):
        assert srv not in self._services, "Duplicated service"
        self._services.add(srv)

    def to_compose_dict(self, topology):
        return get_node_definition(topology, self)


class Broker(BaseNamedModel):
    """Represents a MQTT broker."""

    ENTRY_BROKER = "broker"

    def __init__(self, name, networks, args_compose=None):
        self.networks = networks
        self.args_compose = args_compose
        super().__init__(name)

    @property
    def cmd(self):
        return [self.ENTRY_BROKER]

    def to_compose_dict(self, topology):
        return get_broker_definition(topology, self)


class Network(BaseNamedModel):
    """Represents a network that may contain multiple nodes. 
    Each network has a predefined set of conditions that determine its performance."""

    GATEWAY_PREFIX = "gw"
    assert inflection.underscore(GATEWAY_PREFIX) == GATEWAY_PREFIX
    ENTRY_GATEWAY = "gateway"

    def __init__(
            self, name, netem=None, conditions=None,
            args_compose_net=None, args_compose_gw=None):
        if conditions and conditions not in NetworkConditions:
            raise ValueError("Unexpected conditions value")

        if conditions and not netem:
            netem = NETEM_CONDITIONS[conditions]

        self._netem = netem
        self.args_compose_net = args_compose_net
        self.args_compose_gw = args_compose_gw
        super().__init__(name)

    @property
    def name_gateway(self):
        return "{}_{}".format(self.GATEWAY_PREFIX, self.name)

    @property
    def netem_args(self):
        return self._netem if self._netem else []

    @property
    def cmd_gateway(self):
        return [self.ENTRY_GATEWAY] + list(self.netem_args)

    def to_compose_dict(self, topology):
        return get_network_definition(topology, self)

    def to_gateway_compose_dict(self, topology):
        return get_network_gateway_definition(topology, self)


class TopologyPorts:
    def __init__(
            self, catalogue=_DEFAULT_CATALOGUE, http=_DEFAULT_HTTP,
            ws=_DEFAULT_WS, mqtt=_DEFAULT_MQTT, coap=_DEFAULT_COAP):
        self.catalogue = catalogue
        self.http = http
        self.ws = ws
        self.mqtt = mqtt
        self.coap = coap

    @property
    def config(self):
        return {
            ConfigVars.PORT_CATALOGUE.value: self.catalogue,
            ConfigVars.PORT_HTTP.value: self.http,
            ConfigVars.PORT_WS.value: self.ws,
            ConfigVars.PORT_MQTT.value: self.mqtt,
            ConfigVars.PORT_COAP.value: self.coap
        }


class TopologyRedis:
    WARN_MSG = "Disabled built-in topology Redis service"

    def __init__(self, enabled=True, host=DEFAULT_HOST_REDIS, redis_url=None):
        if not enabled and not redis_url:
            raise ValueError((
                "An explicit Redis URL has to be provided "
                "when the built-in Redis service is disabled"
            ))

        self.enabled = bool(enabled)
        self.host = host
        self.redis_url = redis_url

        if not self.enabled:
            warnings.warn(self.WARN_MSG, Warning)

    @property
    def internal_url(self):
        return f"redis://{self.host}" if self.enabled else self.redis_url

    @property
    def config(self):
        return {ConfigVars.REDIS_URL.value: self.internal_url}

    def to_compose_dict(self, topology):
        return get_redis_definition(topology, self)


class TopologyDockerProxy:
    WARN_MSG = "Disabled built-in topology Docker API Proxy service"

    def __init__(self, enabled=True, host=DEFAULT_HOST_DOCKER_PROXY):
        self.enabled = bool(enabled)
        self.host = host

        if not self.enabled:
            warnings.warn(self.WARN_MSG, Warning)

    @property
    def config(self):
        docker_url = "tcp://{}:2375/".format(self.host)
        return {ConfigVars.DOCKER_PROXY_URL.value: docker_url}

    def to_compose_dict(self, topology):
        return get_docker_proxy_definition(topology)


class Topology:
    """Represents a topology consisting of a set of 
    Nodes and Brokers iterconnected by Networks."""

    def __init__(self, nodes, ports=None, redis=None, docker_proxy=None, brokers=None):
        self.nodes = nodes
        self.ports = ports if ports else TopologyPorts()
        self.redis = redis if redis else TopologyRedis()
        self.docker_proxy = docker_proxy if docker_proxy else TopologyDockerProxy()
        self._brokers = brokers

        assert len(set(self.nodes)) == len(self.nodes)

        assert len(set(self.services).union(self.nodes)) \
            == len([*self.services, *self.nodes])

    @property
    def config(self):
        return {
            **self.ports.config,
            **self.redis.config,
            **self.docker_proxy.config
        }

    @property
    def redis_compose_dict(self):
        return self.redis.to_compose_dict(self)

    @property
    def docker_proxy_compose_dict(self):
        return self.docker_proxy.to_compose_dict(self)

    @property
    def brokers(self):
        brokers = self._brokers if self._brokers else []
        brokers += [node.broker for node in self.nodes if node.broker]
        return list(set(brokers))

    @property
    def networks(self):
        nets_node = set(net for node in self.nodes for net in node.networks)
        nets_brkr = set(net for brk in self.brokers for net in brk.networks)
        return list(nets_node.union(nets_brkr))

    @property
    def services(self):
        return list(set(srv for node in self.nodes for srv in node.services))

    def get_service_nodes(self, service):
        return list(set(node for node in self.nodes for item in node.services if item is service))

    def to_compose_dict(self):
        return get_topology_definition(self)

    def to_compose_yaml(self):
        return yaml.dump(self.to_compose_dict())



import logging
import pprint
import subprocess

import netaddr
from wotemu.utils import (get_current_task, get_network_gateway_task,
                          get_output_iface_for_task, get_task_networks,
                          ping_docker)

_PATH_IPROUTE2_RT_TABLES = "/etc/iproute2/rt_tables"

_logger = logging.getLogger(__name__)
_rtindex = None


def _next_rtable_index():
    def exists_rtable(rtidx):
        cmd = "grep -q '{}\\s.*' /etc/iproute2/rt_tables".format(rtidx)
        return subprocess.run(cmd, shell=True).returncode == 0

    global _rtindex

    if _rtindex is None:
        _rtindex = next(
            idx for idx in reversed(range(10, 201))
            if not exists_rtable(idx))

    return _rtindex


def _rtable_exists(rtable_name):
    route_show_res = subprocess.run(
        "ip route show table {}".format(rtable_name),
        shell=True,
        check=False,
        capture_output=True)

    return route_show_res.returncode == 0


def _build_routing_commands(gw_task, ports_tcp, ports_udp, rtable_name, rtable_mark):
    cmds = []

    rtindex = _next_rtable_index()

    cmds.append("grep -Fxq \"{idx} {name}\" {path} || echo \"{idx} {name}\" >> {path}".format(
        idx=rtindex,
        name=rtable_name,
        path=_PATH_IPROUTE2_RT_TABLES))

    ifname, ifaddr = get_output_iface_for_task(gw_task)

    ifcidr = netaddr.IPNetwork("{}/{}".format(
        ifaddr["addr"], ifaddr["netmask"])).cidr

    cmds.append("ip route add {} via {} onlink dev {} proto kernel src {} table {}".format(
        ifcidr,
        gw_task["EndpointIP"],
        ifname,
        ifaddr["addr"],
        rtable_name))

    cmds.append("ip rule add fwmark {} table {}".format(
        hex(rtable_mark),
        rtable_name))

    def build_iptables_rules(proto, dport):
        return [
            "iptables -A OUTPUT -t mangle -o {} -p {} --dport {} -j MARK --set-mark {}".format(
                ifname,
                proto,
                dport,
                rtable_mark),
            "iptables -A OUTPUT -t mangle -o {} -p {} --sport {} -j MARK --set-mark {}".format(
                ifname,
                proto,
                dport,
                rtable_mark)
        ]

    cmds += [
        rule
        for port in ports_tcp
        for rule in build_iptables_rules("tcp", port)
    ]

    cmds += [
        rule
        for port in ports_udp
        for rule in build_iptables_rules("udp", port)
    ]

    return cmds


def _run_commands(cmds):
    for cmd in cmds:
        _logger.info("# %s", cmd)
        ret = subprocess.run(cmd, shell=True, check=True)
        _logger.info("%s", ret)


def update_routing(conf, rtable_name, rtable_mark, apply):
    ports_tcp = [
        conf.port_http,
        conf.port_ws,
        conf.port_mqtt,
        conf.port_coap
    ]

    ports_udp = [conf.port_coap]

    if conf.other_ports_tcp:
        ports_tcp.extend(conf.other_ports_tcp)

    if conf.other_ports_udp:
        ports_udp.extend(conf.other_ports_udp)

    _logger.info("Updating routes for TCP ports: %s", ports_tcp)
    _logger.info("Updating routes for UDP ports: %s", ports_udp)

    docker_url = conf.docker_proxy_url
    ping_docker(docker_url=docker_url)

    if _rtable_exists(rtable_name=rtable_name):
        _logger.info("Table '%s' exists: Skip configuration", rtable_name)
        return

    task = get_current_task(docker_url=docker_url)
    network_ids = get_task_networks(docker_url=docker_url, task=task)

    gw_tasks = {}

    for net_id in network_ids:
        try:
            the_gw_task = get_network_gateway_task(
                docker_url=docker_url,
                network_id=net_id)
        except:
            _logger.error("Error finding gateway task", exc_info=True)

            raise RuntimeError((
                "Could not find gateway task for network {}. "
                "This was probably due to the fact that "
                "the task was not fully initialized yet."
            ).format(net_id))

        gw_tasks.update({net_id: the_gw_task})

    _logger.debug(
        "Gateway tasks:\n%s",
        pprint.pformat(gw_tasks))

    gw_commands = {
        net_id: _build_routing_commands(
            gw_task=task,
            ports_tcp=ports_tcp,
            ports_udp=ports_udp,
            rtable_name=rtable_name,
            rtable_mark=rtable_mark)
        for net_id, task in gw_tasks.items()
    }

    _logger.info(
        "Routing commands:\n%s",
        pprint.pformat(gw_commands))

    if not apply:
        _logger.warning("Dry run: Skip configuration update")
        return

    for cmds in gw_commands.values():
        _run_commands(cmds)



import asyncio
import importlib
import json
import logging
import os
import pprint
import re
import sys
import time

import docker
import netaddr
import netifaces
import sh
import tornado.httpclient

from wotemu.enums import Labels

_CGROUP_PATH = "/proc/self/cgroup"
_STACK_NAMESPACE = "com.docker.stack.namespace"
_CID_HOST_LEN = 12
_STATE_RUNNING = "running"

_logger = logging.getLogger(__name__)


class NodeHTTPTimeout(Exception):
    pass


async def _ping_http(url):
    http_client = tornado.httpclient.AsyncHTTPClient()

    try:
        await http_client.fetch(url)
        _logger.debug("HTTP ping OK: %s", url)
        return True
    except Exception as ex:
        _logger.debug("HTTP ping Error (%s): %s", url, ex)
        return False
    finally:
        http_client.close()


async def _ping_http_timeout(url, wait, timeout):
    _logger.debug("HTTP ping (%s secs timeout): %s", timeout, url)

    ini = time.time()

    def _raise_timeout():
        if timeout is None:
            return

        diff = time.time() - ini

        if diff >= timeout:
            raise NodeHTTPTimeout(
                "HTTP timeout ({} s): {}".format(timeout, url))

    while True:
        _raise_timeout()

        if (await _ping_http(url)):
            break

        _raise_timeout()
        await asyncio.sleep(wait)


async def wait_node(conf, name, wait=2, timeout=120, find_replicas=True):
    cont_hosts = [name]

    if find_replicas:
        _logger.debug((
            "Attempting to translate service name '%s' "
            "to the container hostnames of all the "
            "replicas for that service"
        ), name)

        try:
            cont_hosts = get_service_container_hostnames(
                docker_url=conf.docker_proxy_url,
                name=name)
        except Exception as ex:
            _logger.warning("Error finding container hostnames: %s", ex)
            _logger.warning("Using untranslated service name: %s", cont_hosts)

    urls = [
        "http://{}:{}".format(host, conf.port_catalogue)
        for host in cont_hosts
    ]

    _logger.debug("Catalogue URLs: %s", urls)

    ping_awaitables = [
        _ping_http_timeout(url=url, wait=wait, timeout=timeout)
        for url in urls
    ]

    await asyncio.gather(*ping_awaitables)


def _find_service_container_hosts(docker_api_client, service_name):
    task_filters = {
        "service": service_name,
        "desired-state": _STATE_RUNNING
    }

    _logger.debug("Filtering Docker tasks using filters: %s", task_filters)

    try:
        service_tasks = docker_api_client.tasks(filters=task_filters)
    except Exception as ex:
        _logger.warning(
            "Error finding Docker tasks (filters: %s): %s",
            task_filters, ex)

        return []

    _logger.debug(
        "Found %s tasks for service: %s",
        len(service_tasks), service_name)

    return [
        task["Status"]["ContainerStatus"]["ContainerID"][:_CID_HOST_LEN]
        for task in service_tasks
    ]


def get_service_container_hostnames(docker_url, name):
    docker_api_client = docker.APIClient(base_url=docker_url)

    _logger.debug("Finding container hostnames for: %s", name)
    service_parts = name.split(".")

    try:
        network_candidate = service_parts[-1]
        docker_api_client.inspect_network(network_candidate)
        _logger.debug("Found network: %s", network_candidate)
        base_name = ".".join(service_parts[:-1])
    except docker.errors.NotFound:
        _logger.debug("Network not found: %s", network_candidate)
        base_name = name

    namespace = get_current_stack_namespace(docker_url)
    service_names = [f"{namespace}_" + base_name]

    if base_name.startswith(f"{namespace}_"):
        service_names.append(base_name)

    ret = [
        _find_service_container_hosts(
            docker_api_client=docker_api_client,
            service_name=service_name)
        for service_name in service_names
    ]

    ret = [host for item in ret for host in item]

    if not len(ret):
        raise Exception("Could not find container hostnames for: %s", name)

    _logger.debug("Service %s container hostnames: %s", name, ret)

    return ret


def ping_docker(docker_url):
    try:
        docker_client = docker.DockerClient(base_url=docker_url)
        docker_client.ping()
    except Exception as ex:
        raise Exception("Could not ping Docker daemon: {}".format(ex))


def get_current_container_id():
    try:
        with open(_CGROUP_PATH, "r") as fh:
            cgroup = fh.read()
    except FileNotFoundError as ex:
        raise Exception((
            "The current environment does not "
            "seem to be a Docker container ({})"
        ).format(ex))

    cid_regex = r"\d+:.+:\/docker\/([a-zA-Z0-9]+)"
    result = re.search(cid_regex, cgroup)

    if not result or len(result.groups()) <= 0:
        _logger.warning("Could not find container ID in:\n%s", cgroup)
        raise Exception("Could not retrieve container ID")

    cid = result.groups()[0]

    _logger.debug("Current container ID: %s", cid)

    return cid


def get_task_container_id(task_dict):
    return task_dict.get("Status", {}).get("ContainerStatus", {}).get("ContainerID", None)


def get_current_task(docker_url):
    docker_api_client = docker.APIClient(base_url=docker_url)
    cid = get_current_container_id()

    task = next((
        task for task in docker_api_client.tasks()
        if get_task_container_id(task) == cid), None)

    if task is None:
        raise Exception("Could not find task for container: {}".format(cid))

    return task


def get_current_stack_namespace(docker_url):
    curr_task = get_current_task(docker_url=docker_url)
    return curr_task.get("Spec", {}).get("ContainerSpec", {}).get("Labels", {}).get(_STACK_NAMESPACE, None)


def get_task_networks(docker_url, task):
    docker_api_client = docker.APIClient(base_url=docker_url)

    network_ids = [
        net["Network"]["ID"]
        for net in task["NetworksAttachments"]
    ]

    networks = {
        net_id: docker_api_client.inspect_network(net_id)
        for net_id in network_ids
    }

    networks = {
        net_id: net_info for net_id, net_info in networks.items()
        if net_info.get("Labels", {}).get(Labels.WOTEMU_NETWORK.value, None) is not None
    }

    return list(networks.keys())


def get_task_labels(docker_url, task_name):
    docker_api_client = docker.APIClient(base_url=docker_url)
    task_info = docker_api_client.inspect_task(task_name)

    return task_info["Spec"]["ContainerSpec"]["Labels"]


def get_network_gateway_task(docker_url, network_id):
    docker_api_client = docker.APIClient(base_url=docker_url)

    network_info = docker_api_client.inspect_network(network_id, verbose=True)

    service_infos = {
        net_name: info
        for net_name, info in network_info["Services"].items()
        if len(net_name) > 0
    }

    _logger.debug(
        "Network %s services:\n%s",
        network_id,
        pprint.pformat(list(service_infos.keys())))

    task_infos = {
        task_info["Name"]: task_info
        for net_name, serv_info in service_infos.items()
        for task_info in serv_info["Tasks"]
    }

    _logger.debug(
        "Network %s tasks:\n%s",
        network_id,
        pprint.pformat(list(task_infos.keys())))

    task_labels = {
        task_name: get_task_labels(docker_url, task_name)
        for task_name in task_infos.keys()
    }

    return next(
        task_infos[task_name]
        for task_name, labels in task_labels.items()
        if labels.get(Labels.WOTEMU_GATEWAY.value, None) is not None)


def get_output_iface_for_task(net_task_dict):
    task_name = net_task_dict["Name"]
    task_addr = netaddr.IPAddress(net_task_dict["EndpointIP"])

    iface_addrs = {
        name: netifaces.ifaddresses(name).get(netifaces.AF_INET)
        for name in netifaces.interfaces()
        if netifaces.ifaddresses(name).get(netifaces.AF_INET)
    }

    _logger.debug(
        "Current container interfaces:\n%s",
        pprint.pformat(iface_addrs))

    ret = next(
        (iface_name, addr)
        for iface_name, iface_addrs in iface_addrs.items()
        for addr in iface_addrs
        if task_addr in netaddr.IPNetwork("{}/{}".format(addr["addr"], addr["netmask"])))

    _logger.debug("Output interface for %s: %s", task_name, ret)

    return ret


def strip_ansi_codes(val):
    """Attribution to: https://stackoverflow.com/a/15780675"""

    return re.sub(r'\x1b\[([0-9,A-Z]{1,2}(;[0-9]{1,2})?(;[0-9]{3})?)?[m|K]?', "", val)


def import_func(module_path, func_name):
    _logger.debug("Attempting to import module: %s", module_path)

    path_root, path_base = os.path.split(module_path)

    if path_root not in sys.path:
        sys.path.insert(0, path_root)

    mod_name, _ext = os.path.splitext(path_base)
    mod_import = importlib.import_module(mod_name)
    mod_dir = dir(mod_import)

    _logger.info("Imported: %s", mod_import)
    _logger.debug("dir(%s): %s", mod_import, mod_dir)

    if func_name not in mod_dir:
        raise Exception("Module {} does not contain function '{}'".format(
            mod_import, func_name))

    return getattr(mod_import, func_name)


async def consume_from_catalogue(wot, port_catalogue, servient_host, thing_id):
    http_client = tornado.httpclient.AsyncHTTPClient()
    cat_url = "http://{}:{}".format(servient_host, port_catalogue)

    _logger.debug("Fetching catalogue: %s", cat_url)

    catalogue_res = await http_client.fetch(cat_url)
    catalogue = json.loads(catalogue_res.body)

    _logger.debug("Catalogue:\n%s", pprint.pformat(catalogue))

    if thing_id not in catalogue:
        raise Exception(f"Thing '{thing_id}' not in catalogue: {cat_url}")

    td_url = "http://{}:{}/{}".format(
        servient_host,
        port_catalogue,
        catalogue[thing_id].strip("/"))

    _logger.debug("Consuming from URL: %s", td_url)

    return await wot.consume_from_url(td_url)


def cgget(name):
    try:
        sh_cgget = sh.Command("cgget")
        cmd_parts = ["-v", "-r", name, "/"]
        proc = sh_cgget(cmd_parts, _err_to_out=True)
        _logger.debug("%s: %s", proc.ran, proc.stdout)
        match = re.search(r"(-?\d+)\n", proc.stdout.decode("utf8"))
        return int(match.group(1)) if match else None
    except:
        _logger.warning("Error running cgget for: %s", exc_info=True)
        return None



import copy
import json
import logging
import socket
import time

import aioredis
import wotemu.config
from wotemu.enums import RedisPrefixes

_logger = logging.getLogger(__name__)
_state = {"redis": None}


async def _get_redis():
    try:
        assert await _state["redis"].ping()
    except:
        _logger.debug("Initializing module-level Redis pool")
        conf = wotemu.config.get_env_config()

        if not conf.redis_url:
            raise RuntimeError("Undefined Redis URL")

        redis_pool = await aioredis.create_redis(conf.redis_url)
        _state["redis"] = redis_pool

    return _state["redis"]


async def write_metric(key, data):
    redis_pool = await _get_redis()
    base_key = socket.getfqdn()

    full_key = "{}:{}:{}:{}".format(
        RedisPrefixes.NAMESPACE.value,
        RedisPrefixes.APP.value,
        base_key,
        key)

    now = time.time()
    data = copy.copy(data)
    data.update({"created_at": now})
    member = json.dumps(data)

    _logger.debug("ZADD %s: %s", full_key, member)

    await redis_pool.zadd(
        key=full_key,
        score=now,
        member=member)



import logging

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())

try:
    import pandas
    pandas.set_option("display.max_columns", None)
except ImportError:
    pass



"""
Reads all the properties from the Thing 
passed as argument in a periodic fashion.
"""

import asyncio
import logging

from wotemu.utils import consume_from_catalogue, wait_node

_logger = logging.getLogger(__name__)


async def _read(consumed_thing, name):
    try:
        _logger.debug("Reading: %s", name)
        val = await consumed_thing.properties[name].read()
        _logger.debug("Read %s = %s", name, val)
    except Exception as ex:
        _logger.warning("Error reading %s: %s", name, ex)


async def app(wot, conf, loop, servient_host, thing_id, interval=5):
    await wait_node(conf=conf, name=servient_host)

    consumed_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=conf.port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    while True:
        for name in consumed_thing.properties:
            await _read(consumed_thing, name)

        await asyncio.sleep(interval)



import asyncio
import copy
import logging
import pprint
import time
import uuid
from functools import wraps

import wotpy.wot.consumed.thing
import wotpy.wot.exposed.thing
from wotpy.protocols.enums import InteractionVerbs

_logger = logging.getLogger(__name__)


def _truncate_key(data_dict, key, max_len=512):
    if not data_dict.get(key):
        return

    str_val = str(data_dict[key])

    if len(str_val) > max_len:
        str_val = f"{str_val[:max_len]} [...]"

    data_dict.update({key: str_val})


class VerbCallback:
    def __init__(self, verb, func, call_args, call_kwargs):
        self.verb = verb
        self.func = func
        self.call_args = call_args
        self.call_kwargs = call_kwargs

    @classmethod
    def json_ex(cls, ex):
        return {
            "type": ex.__class__.__name__,
            "message": str(ex)
        } if ex else None

    def __str__(self):
        return "<{}> {}".format(self.__class__.__name__, self.__dict__)

    @property
    def data(self):
        return {
            "thing": self.thing_id,
            "verb": self.verb,
            "name": self.interaction_name,
            "time": time.time(),
            "host": self.hostname,
            "class": self.thing_class
        }

    @property
    def callback(self):
        return self.thing.deco_cb

    @property
    def loop(self):
        try:
            return asyncio.get_running_loop()
        except:
            return asyncio.get_event_loop()

    @property
    def interaction_name(self):
        if self.call_kwargs.get("name", None) is not None:
            return self.call_kwargs["name"]

        if len(self.call_args) >= 2 and self.call_args[1]:
            return self.call_args[1]

        _logger.warning("Undefined interaction name: %s", self)

        return None

    @property
    def thing(self):
        thing = self.call_args[0]
        assert isinstance(thing, (ExposedThing, ConsumedThing))
        return thing

    @property
    def hostname(self):
        return self.thing.servient.hostname

    @property
    def thing_id(self):
        return self.thing.id

    @property
    def thing_class(self):
        return self.thing.__class__.__name__


class RequestVerbCallback(VerbCallback):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._result = None
        self._error = None
        self._init_time = time.time()
        self.latency = None

    @property
    def data(self):
        data = super().data

        if self.latency is None:
            _logger.warning("Called data property with no latency")

        data.update({
            "latency": self.latency,
            "error": self.json_ex(self.error),
            "result": self.result
        })

        return data

    @property
    def result(self):
        return self._result

    @result.setter
    def result(self, val):
        self._error = None
        self._result = val

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, err):
        self._result = None
        self._error = err

    def update_latency(self):
        self.latency = time.time() - self._init_time

    def _log_item(self, data):
        data_log = copy.copy(data)
        _truncate_key(data_log, "result")

        _logger.log(
            logging.WARNING if self.error else logging.DEBUG,
            "<%s>\n%s",
            self.__class__.__name__,
            pprint.pformat(data_log))

    def create_callback_task(self):
        if not self.callback:
            return

        data = self.data
        self._log_item(data)
        self.loop.create_task(self.callback(data))


class SubscriptionVerbCallback(VerbCallback):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sub_id = uuid.uuid4().hex

    @property
    def data(self):
        data = super().data

        data.update({
            "subscription": self.sub_id
        })

        return data

    def _log_item(self, data):
        data_log = copy.copy(data)
        _truncate_key(data_log, "item")

        _logger.log(
            logging.WARNING if data_log.get("error") else logging.DEBUG,
            "<%s>\n%s",
            self.__class__.__name__,
            pprint.pformat(data_log))

    def decorate_observable(self, obsv):
        if not self.callback:
            return obsv

        def callback_task(data):
            data.update(self.data)
            self._log_item(data)
            self.loop.create_task(self.callback(data))

        event_key = "event"

        def on_next(item):
            try:
                item_data = item.data.__dict__
            except:
                item_data = item.data

            callback_task({
                event_key: "on_next",
                "item": item_data
            })

        def on_error(err):
            callback_task({
                event_key: "on_error",
                "error": self.json_ex(err)
            })

        def on_completed():
            callback_task({
                event_key: "on_completed"
            })

        def on_subscribe():
            callback_task({
                event_key: "on_subscribe"
            })

        def on_finally():
            callback_task({
                event_key: "on_finally"
            })

        obsv = obsv.do_action(
            on_next=on_next,
            on_error=on_error,
            on_completed=on_completed)

        obsv = obsv.do_on_subscribe(on_subscribe)
        obsv = obsv.do_finally(on_finally)

        return obsv


def _request_deco(verb):
    def deco(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            verb_cb = RequestVerbCallback(
                verb=verb,
                func=func,
                call_args=args,
                call_kwargs=kwargs)

            raised_ex = None

            try:
                result = await func(*args, **kwargs)
                verb_cb.result = result
            except Exception as ex:
                raised_ex = ex
                verb_cb.error = raised_ex

            verb_cb.update_latency()
            verb_cb.create_callback_task()

            if raised_ex is not None:
                raise raised_ex
            else:
                return result

        return wrapper

    return deco


def _subscribe_deco(verb):
    def deco(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            verb_cb = SubscriptionVerbCallback(
                verb=verb,
                func=func,
                call_args=args,
                call_kwargs=kwargs)

            obsv = func(*args, **kwargs)

            return verb_cb.decorate_observable(obsv)

        return wrapper

    return deco


class ExposedThing(wotpy.wot.exposed.thing.ExposedThing):
    def __init__(self, *args, **kwargs):
        self.deco_cb = kwargs.pop("deco_cb", None)
        super().__init__(*args, **kwargs)

    @_request_deco(InteractionVerbs.INVOKE_ACTION)
    async def invoke_action(self, *args, **kwargs):
        return await super().invoke_action(*args, **kwargs)

    @_request_deco(InteractionVerbs.WRITE_PROPERTY)
    async def write_property(self, *args, **kwargs):
        return await super().write_property(*args, **kwargs)

    @_request_deco(InteractionVerbs.READ_PROPERTY)
    async def read_property(self, *args, **kwargs):
        return await super().read_property(*args, **kwargs)

    @_subscribe_deco(InteractionVerbs.SUBSCRIBE_EVENT)
    def on_event(self, *args, **kwargs):
        return super().on_event(*args, **kwargs)

    @_subscribe_deco(InteractionVerbs.OBSERVE_PROPERTY)
    def on_property_change(self, *args, **kwargs):
        return super().on_property_change(*args, **kwargs)


class ConsumedThing(wotpy.wot.consumed.thing.ConsumedThing):
    def __init__(self, *args, **kwargs):
        self.deco_cb = kwargs.pop("deco_cb", None)
        super().__init__(*args, **kwargs)

    @_request_deco(InteractionVerbs.INVOKE_ACTION)
    async def invoke_action(self, *args, **kwargs):
        return await super().invoke_action(*args, **kwargs)

    @_request_deco(InteractionVerbs.WRITE_PROPERTY)
    async def write_property(self, *args, **kwargs):
        return await super().write_property(*args, **kwargs)

    @_request_deco(InteractionVerbs.READ_PROPERTY)
    async def read_property(self, *args, **kwargs):
        return await super().read_property(*args, **kwargs)

    @_subscribe_deco(InteractionVerbs.SUBSCRIBE_EVENT)
    def on_event(self, *args, **kwargs):
        return super().on_event(*args, **kwargs)

    @_subscribe_deco(InteractionVerbs.OBSERVE_PROPERTY)
    def on_property_change(self, *args, **kwargs):
        return super().on_property_change(*args, **kwargs)



"""
Sleeps for a while before raising an exception.
"""

import asyncio
import logging
import random
from datetime import datetime

_logger = logging.getLogger(__name__)


async def app(wot, conf, loop, mean_time=300.0):
    mean_time = float(mean_time)
    lambd = 1.0 / mean_time
    sleep_time = random.expovariate(lambd)

    _logger.info(
        "Waiting to raise an error (mean=%s lambda=%s)",
        mean_time, lambd)

    await asyncio.sleep(sleep_time)
    now = datetime.utcnow()
    raise Exception(f"Error raised at {now} UTC")



import logging
import os

import wotpy.wot.servient
import wotpy.wot.td
import wotpy.wot.wot
from wotpy.protocols.http.server import HTTPServer
from wotpy.protocols.mqtt.server import MQTTServer
from wotpy.protocols.ws.server import WebsocketServer

from wotemu.wotpy.things import ConsumedThing, ExposedThing

_logger = logging.getLogger(__name__)


class WoT(wotpy.wot.wot.WoT):
    def __init__(self, *args, **kwargs):
        self.exposed_cb = kwargs.pop("exposed_cb", None)
        self.consumed_cb = kwargs.pop("consumed_cb", None)
        super().__init__(*args, **kwargs)

    def consume(self, td_str):
        td = wotpy.wot.td.ThingDescription(td_str)

        return ConsumedThing(
            servient=self._servient,
            td=td,
            deco_cb=self.consumed_cb)

    def produce(self, model):
        thing = self.thing_from_model(model)

        exposed_thing = ExposedThing(
            servient=self._servient,
            thing=thing,
            deco_cb=self.exposed_cb)

        self._servient.add_exposed_thing(exposed_thing)
        return exposed_thing


def wot_entrypoint(
        port_catalogue=9090, hostname=None, exposed_cb=None, consumed_cb=None,
        port_http=None, port_ws=None, port_coap=None, mqtt_url=None):
    servient = wotpy.wot.servient.Servient(
        hostname=hostname,
        catalogue_port=port_catalogue)

    servers = []

    if port_http:
        _logger.debug("Creating HTTP server on port: %s", port_http)
        servers.append(HTTPServer(port=port_http))

    if port_ws:
        _logger.debug("Creating Websocket server on port: %s", port_ws)
        servers.append(WebsocketServer(port=port_ws))

    if port_coap:
        try:
            from wotpy.protocols.coap.server import CoAPServer
            _logger.debug("Creating CoAP server on port: %s", port_coap)
            servers.append(CoAPServer(port=port_coap))
        except NotImplementedError as ex:
            _logger.warning(ex)

    if mqtt_url:
        _logger.debug("Creating MQTT server on broker: %s", mqtt_url)
        servers.append(MQTTServer(mqtt_url, servient_id=servient.hostname))

    _logger.debug("Adding servers (%s) to servient (%s)", servers, servient)
    [servient.add_server(item) for item in servers]

    return WoT(
        servient=servient,
        exposed_cb=exposed_cb,
        consumed_cb=consumed_cb)



import asyncio
import copy
import json
import logging
import pprint
import time

import aioredis
import docker
import sh
import yaml
from wotemu.cli.utils import find_stack_redis_port, find_stack_services
from wotemu.enums import Labels, RedisPrefixes

_logger = logging.getLogger(__name__)


def _assert_stack_exists(stack):
    services = find_stack_services(stack=stack)

    if len(services) == 0:
        raise Exception(f"No services found for stack: {stack}")


def _get_task_logs(task, tail=30):
    sh_docker = sh.Command("docker")
    cmd_parts = ["service", "logs", "--tail", tail, task["ID"]]

    try:
        _logger.debug("Running: %s", cmd_parts)
        proc = sh_docker(cmd_parts, _err_to_out=True)
    except sh.ErrorReturnCode as ex:
        _logger.warning("Error running %s: %s", cmd_parts, ex)
        return None

    return proc.stdout.decode("utf8")


def _get_stack_snapshot(stack, tail):
    docker_api_client = docker.APIClient()
    services = find_stack_services(stack=stack)

    _logger.info(
        "Inspecting tasks for %s services found on stack: %s",
        len(services), stack)

    return {
        task["ID"]: {
            "task": task,
            "time": time.time(),
            "logs": _get_task_logs(task, tail=tail)
        }
        for service in services
        for task in docker_api_client.tasks(filters={"service": service["ID"]})
    }


def _is_redis(service):
    return service.get("labels", {}).get(Labels.WOTEMU_REDIS.value, None) is not None


async def _zadd_json(redis_url, key, data):
    try:
        redis = await aioredis.create_redis_pool(redis_url)
        score = time.time()
        member = json.dumps(data)
        _logger.debug("ZADD key=%s score=%s", key, score)
        await redis.zadd(key=key, score=score, member=member)
    except:
        _logger.warning("Error on ZADD", exc_info=True)
    finally:
        try:
            redis.close()
            await redis.wait_closed()
        except:
            _logger.warning("Error closing Redis", exc_info=True)


async def _write_snapshot(redis_url, data):
    key = "{}:{}".format(
        RedisPrefixes.NAMESPACE.value,
        RedisPrefixes.SNAPSHOT.value)

    await _zadd_json(redis_url=redis_url, key=key, data=data)


async def _write_compose(redis_url, data):
    key = "{}:{}".format(
        RedisPrefixes.NAMESPACE.value,
        RedisPrefixes.COMPOSE.value)

    await _zadd_json(redis_url=redis_url, key=key, data=data)


def stop_stack(conf, compose_file, stack, redis_url, tail):
    _assert_stack_exists(stack=stack)

    _logger.info("Opening Compose file: %s", compose_file)

    with open(compose_file, "r") as fh:
        compose_content = yaml.load(fh.read(), Loader=yaml.FullLoader)
        compose_content_clone = copy.deepcopy(compose_content)

    _logger.debug("Compose file:\n%s", pprint.pformat(compose_content))

    services = compose_content.get("services", {})

    for service in services.values():
        if not _is_redis(service):
            service.update({"deploy": {"replicas": 0}})

    compose_content["services"] = services

    _logger.debug(
        "Compose file (target):\n%s",
        pprint.pformat(compose_content))

    stack_snapshot = _get_stack_snapshot(stack=stack, tail=tail)

    _logger.info("Updating stack: %s", stack)

    sh_docker = sh.Command("docker")

    proc = sh_docker(
        ["stack", "deploy", "-c", "-", stack],
        _err_to_out=True,
        _in=yaml.dump(compose_content))

    _logger.info("%s\n%s", proc.ran, str(proc).strip())

    if not redis_url:
        redis_port = find_stack_redis_port(stack=stack)
        redis_url = f"redis://127.0.0.1:{redis_port}"

    _logger.info("Using Redis URL: %s", redis_url)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_write_snapshot(redis_url, stack_snapshot))
    loop.run_until_complete(_write_compose(redis_url, compose_content_clone))

    _logger.info("Stack stopped successfully: %s", stack)



import re

import lxml.etree
import pkg_resources


def get_base_template():
    resource_path = "/".join(("templates", "base.html"))
    template = pkg_resources.resource_string(__name__, resource_path)
    tree = lxml.etree.fromstring(template.decode())

    root = next(
        el for el in tree.find("body").iter()
        if el.attrib.get("id") == "root")

    return tree, root


def shorten_task_name(name):
    if not name:
        return name

    return re.sub(
        r"^(.+\.\d+\..{6})(.+)$",
        r"\1",
        name)



"""
Subscribes to all properties and events 
from the Thing passed as argument.
"""

import asyncio
import functools
import logging

from wotemu.utils import consume_from_catalogue, wait_node

_logger = logging.getLogger(__name__)


def _on_next(item, interaction):
    _logger.info("%s: Next\n%s", interaction, item)


def _on_completed(interaction):
    _logger.info("%s: Completed", interaction)


def _on_error(err, interaction, error_event):
    _logger.warning("%s: Error\n%s", interaction, err)
    error_event.set()


def _subscribe(interaction, error_event):
    _logger.debug("Subscribing to: {}".format(interaction))

    on_next = functools.partial(
        _on_next,
        interaction=interaction)

    on_completed = functools.partial(
        _on_completed,
        interaction=interaction)

    on_error = functools.partial(
        _on_error,
        interaction=interaction,
        error_event=error_event)

    return interaction.subscribe(on_next=on_next, on_completed=on_completed, on_error=on_error)


async def _cancel_subs(subs, cancel_sleep=3):
    for sub in subs:
        try:
            _logger.debug("Disposing: {}".format(sub))
            sub.dispose()
        except Exception as ex:
            _logger.debug("Error disposing of %s: %s", sub, ex)

    _logger.info("Waiting %s s. for subscriptions", cancel_sleep)
    await asyncio.sleep(cancel_sleep)


async def _consume_and_subscribe(wot, conf, servient_host, thing_id, error_event):
    await wait_node(conf=conf, name=servient_host)

    consumed_thing = await consume_from_catalogue(
        wot=wot,
        port_catalogue=conf.port_catalogue,
        servient_host=servient_host,
        thing_id=thing_id)

    subs = []

    for name in consumed_thing.properties:
        if consumed_thing.properties[name].observable:
            sub = _subscribe(consumed_thing.properties[name], error_event)
            subs.append(sub)

    for name in consumed_thing.events:
        sub = _subscribe(consumed_thing.events[name], error_event)
        subs.append(sub)

    return subs


async def app(wot, conf, loop, servient_host, thing_id):
    error_event = asyncio.Event()

    consume_subscribe_partial = functools.partial(
        _consume_and_subscribe,
        wot=wot,
        conf=conf,
        servient_host=servient_host,
        thing_id=thing_id,
        error_event=error_event)

    try:
        while True:
            subs = await consume_subscribe_partial()
            await error_event.wait()
            _cancel_subs(subs)
            error_event.clear()
    except asyncio.CancelledError:
        await _cancel_subs(subs)



import logging
import os
import random
import tempfile
import uuid

import pytest
import sh
from wotemu.config import ConfigVars
from wotemu.topology.models import (Broker, Network, Node, NodeApp,
                                    NodeResources, Topology, TopologyPorts,
                                    TopologyRedis)

_logger = logging.getLogger(__name__)


@pytest.fixture
def topology():
    network = Network(
        name="the_net",
        netem=["delay --time 15 --jitter 10 --distribution normal"])

    network_broker = Network(
        name="the_broker_net",
        netem=["delay --time 100 --jitter 50 --distribution normal"])

    node_app = NodeApp(
        path="/root/app.py",
        http=True,
        mqtt=True,
        params={"one": 1, "hello": "world"})

    broker = Broker(
        name="the_broker",
        networks=[network_broker])

    node_resources = NodeResources(
        cpu_limit=0.5,
        mem_limit="100M",
        mem_reservation="50M")

    node = Node(
        name="the_node",
        app=node_app,
        networks=[network],
        broker=broker,
        resources=node_resources)

    return Topology(nodes=[node])


def test_names_underscore(topology):
    node = topology.nodes[0]

    node_kwargs = {
        "app": node.app,
        "networks": node.networks,
        "broker": node.broker
    }

    name_bad = "badName"
    node_bad = Node(name=name_bad, **node_kwargs)

    assert node_bad.name != name_bad

    name_ok = "ok_name"
    node_ok = Node(name=name_ok, **node_kwargs)

    assert node_ok.name == name_ok


def test_topology_compose(topology):
    assert topology.to_compose_dict()


def test_topology_compose_yaml(topology):
    try:
        sh_compose = sh.Command("docker-compose")
    except sh.CommandNotFound:
        pytest.skip("Compose CLI is not available")
        return

    try:
        temp_fh, temp_path = tempfile.mkstemp(suffix=".yml")

        with open(temp_path, "w") as fh:
            fh.write(topology.to_compose_yaml())

        cmd_config = ["-f", temp_path, "config"]
        compose_res = sh_compose(cmd_config, _err_to_out=True)
        assert compose_res
        _logger.debug("Compose validation result: \n%s", compose_res)
    finally:
        try:
            os.close(temp_fh)
            os.remove(temp_path)
        except:
            pass


def test_topology_ports():
    network = Network(name="my_net")
    node_app = NodeApp(path="/root/app.py", http=True)
    node_one = Node(name="node_one", app=node_app, networks=[network])
    node_two = Node(name="node_two", app=node_app, networks=[network])

    port_http = random.randint(10500, 20000)
    topology_ports = TopologyPorts(http=port_http)

    topology = Topology(
        nodes=[node_one, node_two],
        ports=topology_ports)

    compose = topology.to_compose_dict()
    services = compose.get("services")
    env_one = services.get(node_one.name).get("environment")
    env_two = services.get(node_two.name).get("environment")

    assert env_one == env_two
    assert env_one[ConfigVars.PORT_HTTP.value] == str(port_http)
    assert env_one[ConfigVars.PORT_MQTT.value]


def test_topology_redis():
    network = Network(name="my_net")
    node_app = NodeApp(path="/root/app.py", http=True)
    node = Node(name="node", app=node_app, networks=[network])

    redis_host = uuid.uuid4().hex
    redis_url = "redis://{}".format(redis_host)

    top_redis = TopologyRedis(host=redis_host)
    top = Topology(nodes=[node], redis=top_redis)
    compose = top.to_compose_dict()
    services = compose.get("services")
    node_env = services.get(node.name).get("environment")

    assert redis_host in services
    assert node_env[ConfigVars.REDIS_URL.value] == redis_url

    top_redis_disabled = TopologyRedis(
        host=redis_host,
        enabled=False,
        redis_url=redis_url)

    top_disabled = Topology(nodes=[node], redis=top_redis_disabled)
    compose_disabled = top_disabled.to_compose_dict()
    services_disabled = compose_disabled.get("services")
    node_env_disabled = services_disabled.get(node.name).get("environment")

    assert redis_host not in services_disabled
    assert node_env_disabled[ConfigVars.REDIS_URL.value] == redis_url


def test_node_compose(topology):
    node = topology.nodes[0]
    assert node.to_compose_dict(topology)


def test_broker_compose(topology):
    broker = topology.brokers[0]
    assert broker.to_compose_dict(topology)


def test_network_compose(topology):
    network = topology.networks[0]
    assert network.to_compose_dict(topology)


def test_network_gateway_compose(topology):
    network = topology.networks[0]
    assert network.to_gateway_compose_dict(topology)



import io
import logging

import lxml.etree
from wotemu.report.components.base import BaseComponent

_logger = logging.getLogger(__name__)


class FigureBlockComponent(BaseComponent):
    def __init__(self, fig, title=None, height=450, col_class="col", with_row=True):
        self.fig = fig
        self.title = title
        self.height = height
        self.col_class = col_class
        self.with_row = with_row

    def to_element(self):
        try:
            fig_io = io.StringIO()

            self.fig.write_html(
                fig_io,
                include_plotlyjs=False,
                full_html=False)

            fig_html = fig_io.getvalue()

            try:
                el_plot = lxml.etree.fromstring(fig_html)
            except lxml.etree.XMLSyntaxError as ex:
                parser = lxml.etree.HTMLParser()
                _logger.debug("Using %s: %s", parser, ex)
                el_plot = lxml.etree.fromstring(fig_html, parser=parser)

            if self.height:
                el_plot.set("style", f"height: {self.height}px;")

            el_col = lxml.etree.Element(
                "div", attrib={"class": self.col_class})

            if self.title:
                el_title = lxml.etree.Element("h4")
                el_title.text = self.title
                el_col.append(el_title)

            el_col.append(el_plot)

            if not self.with_row:
                return el_col

            el_row = lxml.etree.Element("div", attrib={"class": "row"})
            el_row.append(el_col)

            return el_row
        finally:
            try:
                fig_io.close()
            except:
                pass



import asyncio
import logging
import sys
import time

import aioredis
from hbmqtt.client import MQTTClient

from wotemu.utils import ping_docker

_logger = logging.getLogger(__name__)


async def _ping_redis(redis_url):
    try:
        redis = await aioredis.create_redis_pool(redis_url)
        assert redis.ping()
        return True
    except Exception as ex:
        _logger.debug("Redis unavailable: %s", ex)
        return False
    finally:
        try:
            redis.close()
            await redis.wait_closed()
        except:
            pass


def _ping_docker(docker_url):
    try:
        ping_docker(docker_url)
        return True
    except Exception as ex:
        _logger.debug("Docker API unavailable: %s", ex)
        return False


async def _wait_base(redis_url, docker_url, sleep):
    while not (await _ping_redis(redis_url)):
        await asyncio.sleep(sleep)

    while not _ping_docker(docker_url):
        await asyncio.sleep(sleep)


def _wait_base_timeout(redis_url, docker_url, sleep, timeout):
    loop = asyncio.get_event_loop()
    base_aw = _wait_base(redis_url, docker_url, sleep)
    loop.run_until_complete(asyncio.wait_for(base_aw, timeout=timeout))


async def _ping_broker(broker_url):
    try:
        client = MQTTClient()
        await client.connect(broker_url)
        return True
    except Exception as ex:
        _logger.debug("MQTT broker unavailable: %s", ex)
        return False
    finally:
        try:
            await client.disconnect()
        except:
            pass


async def _wait_broker(broker_url, sleep):
    while not (await _ping_broker(broker_url)):
        await asyncio.sleep(sleep)


def _wait_broker_timeout(broker_url, sleep, timeout):
    loop = asyncio.get_event_loop()
    broker_aw = _wait_broker(broker_url, sleep)
    loop.run_until_complete(asyncio.wait_for(broker_aw, timeout=timeout))


def wait_services(conf, sleep, timeout, waiter_base, waiter_mqtt):
    try:
        if waiter_base:
            _logger.info(
                "Waiting for Docker API (%s) and Redis (%s)",
                conf.docker_proxy_url,
                conf.redis_url)

            _wait_base_timeout(
                redis_url=conf.redis_url,
                docker_url=conf.docker_proxy_url,
                sleep=sleep,
                timeout=timeout)

        if waiter_mqtt and conf.mqtt_url:
            _logger.info("Waiting for MQTT broker (%s)", conf.mqtt_url)

            _wait_broker_timeout(
                broker_url=conf.mqtt_url,
                sleep=sleep,
                timeout=timeout)
        elif waiter_mqtt and not conf.mqtt_url:
            _logger.info("No MQTT broker configuration found in environment")
    except Exception:
        _logger.error("Error waiting for services", exc_info=True)
        sys.exit(1)



import asyncio
import functools
import importlib
import inspect
import logging
import os
import pprint
import re
import signal
import sys
import tempfile

import aioredis
import coloredlogs
import tornado.httpclient
import wotemu.config
import wotemu.wotpy.redis
import wotemu.wotpy.wot
from wotemu.enums import BUILTIN_APPS_MODULES, BuiltinApps
from wotemu.monitor.base import NodeMonitor
from wotemu.utils import (get_current_task, get_network_gateway_task,
                          get_output_iface_for_task, get_task_networks,
                          import_func)

_TIMEOUT = 15
_HTTP_REGEX = r"^https?:\/\/.*"

_logger = logging.getLogger(__name__)


def _fetch_app_file(url):
    _logger.info("Fetching WoT app from: %s", url)

    http_client = tornado.httpclient.HTTPClient()
    res = http_client.fetch(url)
    _os_fh, abs_path = tempfile.mkstemp(suffix=".py")

    _logger.debug("Writing WoT app to: %s", abs_path)

    with open(abs_path, "wb") as fh:
        fh.write(res.body)

    return abs_path


def _remove_tempfile(fpath):
    if not fpath or not fpath.startswith(tempfile.gettempdir()):
        return

    try:
        _logger.debug("Removing tempfile: %s", fpath)
        os.remove(fpath)
    except:
        _logger.warning("Error removing tempfile", exc_info=True)


def _is_builtin_app(app_id):
    try:
        assert app_id in BuiltinApps
        return True
    except:
        pass

    return app_id in [item.value for item in BuiltinApps]


def _import_builtin_app_func(app_id, func_name):
    bapp = next(item for item in BuiltinApps if item.value == app_id)
    mod_name = BUILTIN_APPS_MODULES[bapp]
    mod_import = importlib.import_module(mod_name)
    mod_dir = dir(mod_import)

    _logger.info("Imported: %s", mod_import)
    _logger.debug("dir(%s): %s", mod_import, mod_dir)

    if func_name not in mod_dir:
        raise Exception("Builtin app module {} does not contain {}".format(
            mod_import, func_name))

    return getattr(mod_import, func_name)


def _import_app_func(module_path, func_name):
    _logger.debug("Attempting to import from: %s", module_path)

    if _is_builtin_app(app_id=module_path):
        return _import_builtin_app_func(app_id=module_path, func_name=func_name)

    if re.match(_HTTP_REGEX, module_path):
        module_path = _fetch_app_file(module_path)

    try:
        app_func = import_func(module_path=module_path, func_name=func_name)
        app_func_sig = inspect.signature(app_func)
        _logger.debug("Function (%s) signature: %s", app_func, app_func_sig)
        return app_func
    finally:
        _remove_tempfile(module_path)


def _build_thing_cb(redis_url, loop):
    async def dummy_cb(*args, **kwargs):
        pass

    if not redis_url:
        return dummy_cb, None

    _logger.debug("Creating Redis pool with URL: %s", redis_url)

    redis_pool = loop.run_until_complete(
        aioredis.create_redis_pool(redis_url))

    redis_cb = functools.partial(
        wotemu.wotpy.redis.redis_thing_callback,
        client=redis_pool)

    return redis_cb, redis_pool


def _exception_handler(loop, context):
    level = logging.ERROR if loop.is_running() else logging.DEBUG
    _logger.log(level, "Exception in loop:\n%s", pprint.pformat(context))


async def _start_servient(wot):
    try:
        _logger.debug("Starting Servient: %s", wot.servient)
        await wot.servient.start()
    except Exception:
        _logger.error("Error in Servient startup", exc_info=True)
        sys.exit(1)


async def _stop_servient(wot):
    try:
        _logger.debug("Shutting down Servient: %s", wot.servient)
        await wot.servient.shutdown()
    except Exception:
        _logger.warning("Error in Servient shutdown", exc_info=True)


async def _stop_redis(redis_pool):
    if not redis_pool:
        return

    try:
        _logger.debug("Closing Redis")
        redis_pool.close()
        await redis_pool.wait_closed()
    except Exception:
        _logger.warning("Error in Redis shutdown", exc_info=True)


async def _stop(loop, app_task, wot, redis_pool, monitor, lock):
    if lock.locked():
        _logger.debug("Another stop task is already in progress")
        return

    async with lock:
        try:
            _logger.info("Cancelling WoT app task: %s", app_task)
            app_task.cancel()
            await asyncio.wait_for(app_task, _TIMEOUT)
        except Exception:
            _logger.warning("Error during WoT app cancelation", exc_info=True)

        await _stop_servient(wot=wot)

        if monitor:
            await monitor.stop()

        await _stop_redis(redis_pool=redis_pool)

        _logger.debug("Stopping loop")
        loop.stop()


def _app_done_cb(fut, stop, exit_status):
    if fut.cancelled() or not fut.exception():
        return

    err = repr(fut.exception())
    _logger.error("WoT app error: %s", err)
    exit_status["code"] = 1
    asyncio.ensure_future(stop())


def _get_monitor_ifaces(docker_url):
    curr_task = get_current_task(docker_url=docker_url)

    wotemu_network_ids = get_task_networks(
        docker_url=docker_url,
        task=curr_task)

    gw_tasks = [
        get_network_gateway_task(
            docker_url=docker_url,
            network_id=net_id)
        for net_id in wotemu_network_ids
    ]

    return [
        get_output_iface_for_task(gw_task)[0]
        for gw_task in gw_tasks
    ]


def run_app(
        conf, path, func, func_param, hostname,
        enable_http, enable_mqtt, enable_coap, enable_ws, disable_monitor):
    if not enable_http and not enable_mqtt and not enable_coap and not enable_ws:
        _logger.warning("No protocol bindings have been enabled")

    port_http = conf.port_http if enable_http else None
    port_ws = conf.port_ws if enable_ws else None
    port_coap = conf.port_coap if enable_coap else None
    mqtt_url = conf.mqtt_url if enable_mqtt else None

    if enable_mqtt and not mqtt_url:
        _logger.warning("MQTT is enabled but broker is undefined")

    app_func = _import_app_func(module_path=path, func_name=func)

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(_exception_handler)

    thing_cb, redis_pool = _build_thing_cb(
        redis_url=conf.redis_url,
        loop=loop)

    wot_kwargs = {
        "port_catalogue": conf.port_catalogue,
        "exposed_cb": thing_cb,
        "consumed_cb": thing_cb,
        "port_http": port_http,
        "port_ws": port_ws,
        "port_coap": port_coap,
        "mqtt_url": mqtt_url,
        "hostname": hostname
    }

    _logger.debug("Building WoT entrypoint with args: %s", wot_kwargs)
    wot = wotemu.wotpy.wot.wot_entrypoint(**wot_kwargs)

    asyncio.ensure_future(_start_servient(wot))

    app_args = (wot, conf, loop)
    app_kwargs = {key: val for key, val in func_param}

    _logger.debug(
        "WoT app call signature (positional and keyword):\n%s\n%s",
        pprint.pformat(app_args),
        pprint.pformat(app_kwargs))

    _logger.info("Scheduling task for WoT app: %s", app_func)
    app_task = asyncio.ensure_future(app_func(*app_args, **app_kwargs))

    monitor = None

    if not disable_monitor:
        _logger.info("Scheduling startup task for node monitor")
        ifaces = _get_monitor_ifaces(docker_url=conf.docker_proxy_url)
        monitor = NodeMonitor(redis_url=conf.redis_url, packet_ifaces=ifaces)
        asyncio.ensure_future(monitor.start())

    stop = functools.partial(
        _stop,
        loop=loop,
        app_task=app_task,
        wot=wot,
        redis_pool=redis_pool,
        monitor=monitor,
        lock=asyncio.Lock())

    exit_status = {}

    app_done_cb = functools.partial(
        _app_done_cb,
        stop=stop,
        exit_status=exit_status)

    app_task.add_done_callback(app_done_cb)

    def sig_handler():
        _logger.debug("Received stop signal")
        asyncio.ensure_future(stop())

    for name in {"SIGINT", "SIGTERM"}:
        loop.add_signal_handler(getattr(signal, name), sig_handler)

    try:
        _logger.debug("Running loop")
        loop.run_forever()
    finally:
        _logger.debug("Closing loop")
        loop.close()
        exit_code = exit_status.get("code")

        if exit_code:
            _logger.warning("Exiting with code: %s", exit_code)
            exit(exit_code)



import logging

import lxml.etree
import yaml
from wotemu.report.components.base import BaseComponent

_logger = logging.getLogger(__name__)


class CodeBlockComponent(BaseComponent):
    def __init__(self, content, title=None, subtitle=None):
        self.content = content
        self.title = title
        self.subtitle = subtitle

    def to_element(self):
        card = lxml.etree.Element("div", attrib={"class": "card"})
        card_body = lxml.etree.Element("div", attrib={"class": "card-body"})

        card_pre = lxml.etree.Element(
            "pre", attrib={"class": "mb-0 text-muted"})

        card.append(card_body)

        if self.title:
            card_title = lxml.etree.Element(
                "h5", attrib={"class": "card-title"})

            card_title.text = self.title
            card_body.append(card_title)

        if self.subtitle:
            card_subtitle = lxml.etree.Element(
                "h6", attrib={"class": "card-subtitle mb-2"})

            card_subtitle.text = self.subtitle
            card_body.append(card_subtitle)

        card_body.append(card_pre)
        card_pre.text = yaml.dump(self.content)

        return card



"""Packet monitoring task. 

There is a known issue in this module.

The multiprocessing module starts a "resource tracker" process 
when using the spawn context, as described in the Python docs:

https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods

A Pyshark LiveCapture may be stopped gracefully by raising a 
StopCapture exception inside the packet handling callback. 
However, that callback may not be called for a while if no new packets arrive. 
In this case the only option is to terminate() the Pyshark process and 
capture the signal, manually killing the dumpcap and tshark child processes. 
This causes the "resource tracker" process to end up as a zombie.
"""

import asyncio
import functools
import logging
import multiprocessing
import os
import queue
import signal
import time

import pyshark
from pyshark.capture.capture import StopCapture

_logger = logging.getLogger(__name__)


def _build_display_filter(conf):
    return (
        "tcp.port=={port_mqtt} or "
        "tcp.port=={port_http} or "
        "tcp.port=={port_ws} or "
        "tcp.port=={port_coap} or "
        "udp.port=={port_coap}"
    ).format(
        port_http=conf.port_http,
        port_ws=conf.port_ws,
        port_mqtt=conf.port_mqtt,
        port_coap=conf.port_coap)


def _packet_to_dict(packet):
    ret = {
        "len": int(packet.length),
        "src": packet.ip.src,
        "dst": packet.ip.dst,
        "proto": packet.layers[-1].layer_name,
        "transport": packet.layers[2].layer_name,
        "time": float(packet.sniff_timestamp)
    }

    try:
        ret.update({
            "srcport": int(packet.tcp.srcport),
            "dstport": int(packet.tcp.dstport)
        })
    except AttributeError:
        pass

    try:
        ret.update({
            "srcport": int(packet.udp.srcport),
            "dstport": int(packet.udp.dstport)
        })
    except AttributeError:
        pass

    return ret


def _packet_callback(packet, output_queue, stop_event):
    packet_dict = _packet_to_dict(packet)

    try:
        output_queue.put_nowait(packet_dict)
    except queue.Full:
        pass

    if stop_event.is_set():
        raise StopCapture()


def _start_capture(display_filter, interface, output_queue, stop_event):
    capture = pyshark.LiveCapture(
        interface=interface,
        only_summaries=False,
        display_filter=display_filter)

    def stop_capture(signum, frame):
        try:
            # Accessing the private interface of LiveCapture is
            # less than ideal, but I fail to see another option.
            for proc in capture._running_processes:
                os.kill(proc.pid, signal.SIGKILL)
        except:
            pass

    signal.signal(signal.SIGINT, stop_capture)
    signal.signal(signal.SIGTERM, stop_capture)

    on_packet = functools.partial(
        _packet_callback,
        output_queue=output_queue,
        stop_event=stop_event)

    capture.apply_on_packets(on_packet)


async def _process_queue(output_queue, queue_size, async_cb):
    items = []

    try:
        counter = 0
        while not output_queue.empty():
            item = output_queue.get_nowait()
            items.append(item)
            counter += 1
            if counter > queue_size:
                break
    except queue.Empty:
        pass

    if len(items) > 0:
        await async_cb(items)


def _check_proc_health(proc):
    is_alive = proc.is_alive()
    exitcode = proc.exitcode

    _logger.debug(
        "Checking packet monitor health (is_alive=%s) (exitcode=%s)",
        is_alive, exitcode)

    if not is_alive or exitcode is not None:
        err_msg = (
            "Tshark packet capture process "
            "stopped prematurely "
            "(exit code: {})"
        ).format(proc.exitcode)

        raise RuntimeError(err_msg)


async def _wait_process_exit(proc, timeout, sleep=1):
    _logger.debug("Waiting %s secs for exit: %s", timeout, proc)

    time_limit = time.time() + timeout

    while time.time() <= time_limit:
        if proc.exitcode is not None:
            return

        await asyncio.sleep(sleep)


async def _terminate_process(proc, stop_event, stop_sleep, stop_timeout, stop_join_timeout):
    _logger.debug("Setting stop event for: %s", proc)
    stop_event.set()

    await _wait_process_exit(proc, stop_timeout, sleep=stop_sleep)

    if proc.exitcode is not None:
        _logger.debug("%s exited with code: %s", proc, proc.exitcode)
        return

    _logger.warning("%s did not terminate with stop event", proc)

    try:
        _logger.info("Terminating process: %s", proc)
        proc.terminate()
        await _wait_process_exit(proc, stop_join_timeout, sleep=stop_sleep)
    except:
        _logger.warning("Error terminating", exc_info=True)
    finally:
        try:
            if proc.exitcode is None:
                _logger.warning("Killing process: %s", proc)
                proc.kill()
        except:
            _logger.warning("Error killing", exc_info=True)


async def monitor_packets(
        conf, interface, async_cb,
        queue_size=100, sleep=5.0,
        stop_sleep=0.1, stop_timeout=5.0, stop_join_timeout=5.0):
    spawn_ctx = multiprocessing.get_context("spawn")
    display_filter = _build_display_filter(conf=conf)
    output_queue = spawn_ctx.Queue(queue_size)
    stop_event = spawn_ctx.Event()

    proc_target = functools.partial(
        _start_capture,
        display_filter=display_filter,
        interface=interface,
        output_queue=output_queue,
        stop_event=stop_event)

    _logger.debug(
        "Starting LiveCapture on interface %s with display filter: %s",
        interface,
        display_filter)

    proc = spawn_ctx.Process(
        target=proc_target,
        daemon=True,
        name=f"PysharkCapture-{interface}")

    proc.start()

    process_queue = functools.partial(
        _process_queue,
        output_queue=output_queue,
        queue_size=queue_size,
        async_cb=async_cb)

    check_proc_health = functools.partial(
        _check_proc_health,
        proc=proc)

    terminate_process = functools.partial(
        _terminate_process,
        proc=proc,
        stop_event=stop_event,
        stop_sleep=stop_sleep,
        stop_timeout=stop_timeout,
        stop_join_timeout=stop_join_timeout)

    try:
        while True:
            await process_queue()
            check_proc_health()
            await asyncio.sleep(sleep)
    except asyncio.CancelledError:
        _logger.debug("Cancelled Task for: %s", proc)
    finally:
        if proc.exitcode is None:
            await terminate_process()



"""
Exposes a Thing with an action that 
can be invoked to stress the CPU. 
"""

import asyncio
import json
import logging
import pprint
import time

from wotpy.wot.enums import DataType

_logger = logging.getLogger(__name__)

_DESCRIPTION = {
    "id": "urn:org:fundacionctic:thing:worker",
    "name": "Dummy worker Thing",
    "actions": {
        "doWork": {
            "safe": True,
            "idempotent": False,
            "input": {
                "type": DataType.OBJECT
            },
            "output": {
                "type": DataType.OBJECT
            }
        }
    }
}


async def _action_handler(params):
    params = params or {}
    params_input = params.get("input", {}) or {}

    time_step = float(params_input.get("step", 0.5))
    time_total = float(params_input.get("total", 3.0))
    time_sleep = float(params_input.get("sleep", 0.1))

    stamp_end = time.time() + time_total

    while time.time() < stamp_end:
        stamp_step_end = time.time() + time_step

        while time.time() < stamp_step_end:
            time.time() ** 2

        await asyncio.sleep(time_sleep)

    return {"ret": params_input.get("return", time.time())}


async def app(wot, conf, loop):
    _logger.info(
        "Producing Thing:\n%s",
        pprint.pformat(_DESCRIPTION))

    exposed_thing = wot.produce(json.dumps(_DESCRIPTION))
    exposed_thing.set_action_handler("doWork", _action_handler)
    exposed_thing.expose()



import asyncio
import functools
import json
import logging
import os
import platform
import pprint
import re
import socket
import time

import docker
import psutil
import sh
import wotemu.config
from wotemu.topology.compose import ENV_KEY_SERVICE_NAME
from wotemu.utils import (cgget, get_current_container_id, get_current_task,
                          get_task_networks)

_CPU_USAGE = "cpuacct.usage"
_CPU_PERIOD = "cpu.cfs_period_us"
_CPU_QUOTA = "cpu.cfs_quota_us"
_MEM_LIMIT = "memory.limit_in_bytes"
_MEM_USAGE = "memory.usage_in_bytes"
_NSLOOKUP_REGEX = r"Name:\s.+\..+\nAddress:\s(.+)"
_LSCPU_REGEX = r"Model\sname:[\s\t]*(.+)"
_UNKNOWN_CPU = "Unknown CPU"

_logger = logging.getLogger(__name__)

_times = {
    "system": None,
    "cgroup": None
}

_cache = {}


class CggetError(RuntimeError):
    def __init__(self, key):
        msg = (
            "Could not read cgroup parameter '{}'"
            "using the cgget command. Please make "
            "sure that cgget is installed on your system."
        ).format(key)

        super().__init__(msg)


def _get_cpu_nanos_system():
    cput = psutil.cpu_times()
    return (cput.user + cput.system + cput.idle) * 1e9


def _get_cpu_nanos():
    usage = cgget(_CPU_USAGE)

    if usage is None:
        raise CggetError(_CPU_USAGE)

    return usage


def _is_times_state_empty():
    return _times["system"] is None or _times["cgroup"] is None


def _update_times_state():
    _times.update({
        "system": _get_cpu_nanos_system(),
        "cgroup": _get_cpu_nanos()
    })


def _get_cpu_percent():
    if _is_times_state_empty():
        _update_times_state()
        return None

    curr_cpu = _get_cpu_nanos()
    curr_sys = _get_cpu_nanos_system()

    delta_cpu = float(curr_cpu - _times["cgroup"])
    delta_sys = float(curr_sys - _times["system"])

    _times.update({
        "system": curr_sys,
        "cgroup": curr_cpu
    })

    if delta_cpu > 0 and delta_sys > 0:
        return (delta_cpu / delta_sys) * psutil.cpu_count() * 100.0
    else:
        return 0.0


def _cgget_cpu_period(use_cache=True):
    if use_cache and _cache.get(_CPU_PERIOD, None) is not None:
        return _cache[_CPU_PERIOD]

    period = cgget(_CPU_PERIOD)

    if period is None:
        raise CggetError(_CPU_PERIOD)

    _cache[_CPU_PERIOD] = period

    return period


def _cgget_cpu_quota(use_cache=True):
    if use_cache and _cache.get(_CPU_QUOTA, None) is not None:
        return _cache[_CPU_QUOTA]

    quota = cgget(_CPU_QUOTA)

    if quota is None:
        raise CggetError(_CPU_QUOTA)

    _cache[_CPU_QUOTA] = quota

    return quota


def _get_cpu_constraint():
    period = _cgget_cpu_period(use_cache=True)
    quota = _cgget_cpu_quota(use_cache=True)
    return (float(quota) / float(period)) if quota > 0 else None


def _get_cpu():
    cpu_percent = _get_cpu_percent()

    if cpu_percent is None:
        return None

    ret = {"cpu_percent": round(cpu_percent, 1)}

    cpu_constraint = _get_cpu_constraint()

    if cpu_constraint:
        constraint_ratio = (cpu_percent * 1e-2) / cpu_constraint
        constraint_percent = round(constraint_ratio * 1e2, 1)
        ret.update({"cpu_percent_constraint": constraint_percent})

    return ret


def _cgget_mem_limit(use_cache=True):
    if use_cache and _cache.get(_MEM_LIMIT, None) is not None:
        return _cache[_MEM_LIMIT]

    limit_bytes = cgget(_MEM_LIMIT)
    _cache[_MEM_LIMIT] = limit_bytes

    return limit_bytes


def _get_memory_limit():
    limit_bytes = _cgget_mem_limit(use_cache=True)
    total_bytes_system = psutil.virtual_memory().total

    if limit_bytes and limit_bytes < total_bytes_system:
        return limit_bytes
    else:
        return total_bytes_system


def _get_memory():
    mem_usage = cgget(_MEM_USAGE)

    if mem_usage is None:
        raise CggetError(_MEM_USAGE)

    mem_limit = _get_memory_limit()
    mem_usage_mb = mem_usage / (1024.0 ** 2)
    mem_limit_mb = mem_limit / (1024.0 ** 2)
    mem_percent = (mem_usage_mb / mem_limit_mb) * 100.0

    return {
        "mem_mb": round(mem_usage_mb, 2),
        "mem_percent": round(mem_percent, 1)
    }


def _read_system(data, read_funcs):
    datum = {"time": time.time()}

    read_results = []

    for func in read_funcs:
        try:
            read_results.append(func())
        except Exception as ex:
            _logger.warning("System monitor error: %s", repr(ex))

    [
        datum.update(item)
        for item in read_results if item
    ]

    data.append(datum)


async def _invoke_callback(data, group_size, async_cb):
    if not group_size or len(data) >= group_size:
        await async_cb(data)
        data.clear()


async def monitor_system(async_cb, sleep=5.0, group_size=2):
    data = []

    read_system = functools.partial(
        _read_system,
        data=data,
        read_funcs=(_get_cpu, _get_memory))

    invoke_callback = functools.partial(
        _invoke_callback,
        data=data,
        group_size=group_size,
        async_cb=async_cb)

    try:
        _get_cpu()
        await asyncio.sleep(sleep)

        while True:
            read_system()
            await invoke_callback()
            await asyncio.sleep(sleep)
    except asyncio.CancelledError:
        _logger.debug("Cancelled system usage task")


def _inspect_current_networks():
    conf = wotemu.config.get_env_config()
    docker_url = conf.docker_proxy_url
    curr_task = get_current_task(docker_url=docker_url)
    net_ids = get_task_networks(docker_url=docker_url, task=curr_task)
    docker_api_client = docker.APIClient(base_url=docker_url)

    return [
        docker_api_client.inspect_network(nid)
        for nid in net_ids
    ]


def _get_subnets_cidr():
    networks_data = _inspect_current_networks()

    return {
        net["Name"]: [
            item["Subnet"]
            for item in net.get("IPAM", {}).get("Config", [])
        ]
        for net in networks_data
    }


def _get_service_vips():
    assert ENV_KEY_SERVICE_NAME in os.environ
    service_name = os.environ[ENV_KEY_SERVICE_NAME]

    networks_data = _inspect_current_networks()

    net_names = [
        net["Name"]
        for net in networks_data
    ]

    nslookup = sh.Command("nslookup")

    nslookup_results = {
        net_name: nslookup(f"{service_name}.{net_name}")
        for net_name in net_names
    }

    _logger.debug(
        "VIP nslookup results for service '%s':\n%s",
        service_name,
        pprint.pformat(nslookup_results))

    re_results = {
        net_name: re.search(_NSLOOKUP_REGEX, str(res))
        for net_name, res in nslookup_results.items()
    }

    assert all(re_results.values())
    assert all(len(res.groups()) == 1 for res in re_results.values())

    vips = {
        net_name: res.groups()[0]
        for net_name, res in re_results.items()
    }

    _logger.debug(
        "Service '%s' VIPs obtained from nslookup: %s",
        service_name, vips)

    return vips


def _get_task_id():
    conf = wotemu.config.get_env_config()
    docker_url = conf.docker_proxy_url
    curr_task = get_current_task(docker_url=docker_url)
    return curr_task["ID"]


def _get_cpu_model_lscpu():
    lscpu = sh.Command("lscpu")
    result = lscpu()
    match = re.search(_LSCPU_REGEX, str(result))
    assert match
    assert len(match.groups()) == 1
    return match.groups()[0]


def _get_cpu_model_psutil():
    arch = platform.processor()
    cores = psutil.cpu_count(logical=False)
    freq = psutil.cpu_freq().max or psutil.cpu_freq().current
    freq = round(freq / 1024.0, 1)
    return "CPU {} {} cores {}GHz".format(arch, cores, freq)


def _get_cpu_model():
    try:
        return _get_cpu_model_lscpu()
    except:
        _logger.warning("Error reading CPU model", exc_info=True)

    try:
        return _get_cpu_model_psutil()
    except:
        _logger.warning("Error reading CPU model", exc_info=True)

    return _UNKNOWN_CPU


def _get_constraints():
    mem_limit_bytes = _get_memory_limit()
    mem_limit_mb = round(mem_limit_bytes / (1024.0 ** 2), 3)
    ret = {"mem_limit_mb": mem_limit_mb}
    cpu_constraint = _get_cpu_constraint()

    if cpu_constraint:
        cpu_percent = round(cpu_constraint * 1e2, 2)
        ret.update({"cpu_percent": cpu_percent})

    return ret


def get_node_info():
    net_addrs = {
        key: [item._asdict() for item in val]
        for key, val in psutil.net_if_addrs().items()
    }

    procs = {
        proc.pid: proc.info
        for proc in psutil.process_iter(["name", "username"])
    }

    info = {
        "cpu_count": psutil.cpu_count(),
        "cpu_model": _get_cpu_model(),
        "mem_total": psutil.virtual_memory().total,
        "net": net_addrs,
        "python_version": platform.python_version(),
        "uname": platform.uname()._asdict(),
        "process": procs,
        "boot_time": psutil.boot_time(),
        "env": dict(os.environ),
        "hostname": socket.gethostname()
    }

    try:
        info.update({"service_vips": _get_service_vips()})
    except:
        _logger.warning("Error reading service VIPs", exc_info=True)

    try:
        info.update({"networks_cidr": _get_subnets_cidr()})
    except:
        _logger.warning("Error reading subnetworks CIDR", exc_info=True)

    try:
        info.update({"container_id": get_current_container_id()})
    except:
        _logger.warning("Error reading container ID", exc_info=True)

    try:
        info.update({"task_id": _get_task_id()})
    except:
        _logger.warning("Error reading task ID", exc_info=True)

    try:
        info.update({"constraints": _get_constraints()})
    except:
        _logger.warning("Error reading constraints", exc_info=True)

    return json.loads(json.dumps(info))



import copy
import datetime
import functools
import json
import logging
import math
import os
import time

import lxml.etree
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import wotpy.wot.consumed.thing
import wotpy.wot.exposed.thing
from plotly.subplots import make_subplots
from wotemu.report.components.code_block import CodeBlockComponent
from wotemu.report.components.container import ContainerComponent
from wotemu.report.components.figure_block import FigureBlockComponent
from wotemu.report.components.task_list import TaskListComponent
from wotemu.report.components.task_section import TaskSectionComponent
from wotemu.report.utils import shorten_task_name
from wotpy.protocols.enums import InteractionVerbs

_MIN_HEIGHT = 400

_logger = logging.getLogger(__name__)


class CacheMiss(Exception):
    pass


class ReportBuilder:
    def __init__(self, reader, use_cache=True):
        self._reader = reader
        self._cache = {}
        self.use_cache = use_cache

    def _serialize_params(self, args, kwargs):
        return (args, frozenset(kwargs.items()))

    def _get_copy(self, result):
        try:
            assert isinstance(result, (pd.DataFrame, pd.Series))
            return result.copy(deep=True)
        except:
            return copy.deepcopy(result)

    def _cache_get(self, func, args, kwargs):
        params_key = self._serialize_params(args, kwargs)

        if func not in self._cache or params_key not in self._cache[func]:
            raise CacheMiss

        return self._cache[func][params_key]

    def _cache_set(self, func, args, kwargs, result):
        params_key = self._serialize_params(args, kwargs)
        self._cache[func] = self._cache.get(func, {})
        self._cache[func][params_key] = result

    async def _reader_exec(self, func, *args, **kwargs):
        if not self.use_cache:
            return await func(*args, **kwargs)

        try:
            result = self._cache_get(func, args, kwargs)
        except CacheMiss:
            result = await func(*args, **kwargs)
            self._cache_set(func, args, kwargs, result)

        return self._get_copy(result)

    async def _get_system_df(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_system_df,
            *args, **kwargs)

    async def _get_packet_df(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_packet_df,
            *args, **kwargs)

    async def _get_service_traffic_df(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_service_traffic_df,
            *args, **kwargs)

    async def _get_thing_df(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_thing_df,
            *args, **kwargs)

    async def _get_tasks(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_tasks,
            *args, **kwargs)

    async def _get_info_map(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_info_map,
            *args, **kwargs)

    async def _get_info(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_info,
            *args, **kwargs)

    async def _get_snapshot_df(self, *args, **kwargs):
        return await self._reader_exec(
            self._reader.get_snapshot_df,
            *args, **kwargs)

    def reset_cache(self):
        self._cache = {}

    async def build_task_mem_figure(self, task):
        df_system = await self._get_system_df(task=task)

        if df_system.empty:
            return None

        df_system = df_system.reset_index()

        fig = make_subplots(specs=[[{"secondary_y": True}]])

        trace_primary = go.Scatter(
            x=df_system["date"],
            y=df_system["mem_mb"],
            name="Memory (MB)")

        fig.add_trace(trace_primary, secondary_y=False)

        trace_secondary = go.Scatter(
            x=df_system["date"],
            y=df_system["mem_percent"],
            name="Memory (%)",
            line=dict(dash="dot"))

        fig.add_trace(trace_secondary, secondary_y=True)

        fig.update_layout(title_text="Memory usage")
        fig.update_xaxes(title_text="Date (UTC)")
        fig.update_yaxes(title_text="MB", secondary_y=False)
        fig.update_yaxes(title_text="%", secondary_y=True)

        return fig

    async def build_task_cpu_figure(self, task):
        df_system = await self._get_system_df(task=task)

        if df_system.empty:
            return None

        df_system = df_system.reset_index()

        has_secondary = "cpu_percent_constraint" in df_system \
            and df_system["cpu_percent_constraint"].any()

        fig = make_subplots(specs=[[{"secondary_y": has_secondary}]])

        trace_primary = go.Scatter(
            x=df_system["date"],
            y=df_system["cpu_percent"],
            name="CPU total (%)")

        fig.add_trace(trace_primary, secondary_y=False)

        fig.update_layout(title_text="CPU usage")
        fig.update_xaxes(title_text="Date (UTC)")

        fig.update_yaxes(
            title_text="% over total",
            secondary_y=False)

        if has_secondary:
            trace_secondary = go.Scatter(
                x=df_system["date"],
                y=df_system["cpu_percent_constraint"],
                name="CPU constraint (%)",
                line=dict(dash="dot"))

            fig.add_trace(trace_secondary, secondary_y=True)

            fig.update_yaxes(
                title_text="% over constraint",
                secondary_y=True)

        return fig

    async def _build_task_packet_figure(self, task, freq, col):
        df = await self._get_packet_df(task=task, extended=True)

        if df is None or df.empty:
            return None

        df = df.reset_index()
        df = df[df["dstport"].notna() & df["srcport"].notna()]

        col_series = {}
        grouper = pd.Grouper(freq=freq)

        for key in df[col].unique():
            df_key = df[df[col] == key]
            df_key.set_index(["date"], inplace=True)
            ser = df_key["len"].groupby(grouper).sum() / 1024.0
            col_series[key] = ser

        iface_traces = {
            key: go.Scatter(x=ser.index, y=ser, name=key)
            for key, ser in col_series.items()
        }

        fig = make_subplots()
        [fig.add_trace(trace) for trace in iface_traces.values()]
        fig.update_layout(barmode="stack", showlegend=True)
        fig.update_xaxes(title_text="Date (UTC)")
        fig.update_yaxes(title_text="KB")

        return fig

    async def build_task_packet_iface_figure(self, task, freq="10s"):
        fig = await self._build_task_packet_figure(
            task=task,
            freq=freq,
            col="iface")

        if not fig:
            return None

        fig.update_layout(
            title_text="Task data transfer by interface ({} windows)".format(freq))

        return fig

    async def build_task_packet_protocol_figure(self, task, freq="10s"):
        fig = await self._build_task_packet_figure(
            task=task,
            freq=freq,
            col="proto")

        if not fig:
            return None

        fig.update_layout(
            title_text="Task data transfer by protocol ({} windows)".format(freq))

        return fig

    def _get_traffic(self, df, col_service, col_task, service, task):
        df_where = df[
            (df[col_service] == service) &
            (df[col_task] == task)]

        assert len(df_where) in [0, 1]

        return 0 if df_where.empty else (df_where.iloc[0]["len"] / 1024.0)

    async def build_service_traffic_figure(self, inbound, colorscale="Portland", height_task=70):
        df = await self._get_service_traffic_df(inbound=inbound)

        if df.empty:
            return None

        col_service = "dst_service" if inbound else "src_service"
        col_task = "src_task" if inbound else "dst_task"

        heatmap_x = sorted(list(df[col_service].unique()))
        heatmap_y = sorted(list(df[col_task].unique()))

        get_traffic_partial = functools.partial(
            self._get_traffic,
            df=df,
            col_service=col_service,
            col_task=col_task)

        heatmap_z = [
            [
                get_traffic_partial(service=service, task=task)
                for service in heatmap_x
            ]
            for task in heatmap_y
        ]

        fig = make_subplots()

        trace = go.Heatmap(
            colorscale=colorscale,
            z=heatmap_z,
            x=[shorten_task_name(item) for item in heatmap_x],
            y=[shorten_task_name(item) for item in heatmap_y])

        fig.add_trace(trace)

        title = "{} traffic (KB) (service {})".format(
            "Task - service" if inbound else "Service - task",
            "inbound" if inbound else "outbound")

        height = height_task * len(heatmap_y)
        height = max(height, _MIN_HEIGHT)

        fig.update_xaxes(title_text="Service")
        fig.update_yaxes(title_text="Task")
        fig.update_layout(title_text=title, height=height)

        return fig

    async def build_network_traffic_figure(self, freq="10s", height=500):
        tasks = await self._get_tasks()

        df_packets = [
            await self._get_packet_df(task=task, extended=True)
            for task in tasks
        ]

        df = pd.concat([item for item in df_packets if item is not None])

        if df.empty:
            return None

        df = df.groupby([
            pd.Grouper(freq=freq, level="date"),
            pd.Grouper("network")
        ]).sum()

        df.reset_index(inplace=True)

        df["len_kb"] = df["len"] / 1024.0

        fig = px.line(
            df,
            x="date",
            y="len_kb",
            color="network",
            height=height,
            color_discrete_sequence=px.colors.qualitative.Dark24)

        fig.update_xaxes(title_text="Date (UTC)")
        fig.update_yaxes(title_text="KB")

        fig.update_layout(
            showlegend=True,
            title_text="Data transfer by network ({} windows)".format(freq))

        return fig

    async def build_thing_counts_figure(self, task, facet_col_wrap=2):
        df = await self._get_thing_df(task=task)

        if df.empty:
            return None

        df.reset_index(inplace=True)

        df["error_sub"] = (df["verb"] == "subscribeevent") \
            & (df["event"] == "on_error") \
            if "event" in df else False

        df["error_req"] = (df["verb"] != "subscribeevent") \
            & (df["error"].notna()) \
            if "error" in df else False

        cat_ok = "OK"
        cat_er = "Error"

        df["success"] = df["error_req"] | df["error_sub"]
        df["success"].replace(False, cat_ok, inplace=True)
        df["success"].replace(True, cat_er, inplace=True)
        df["thing"] = df["thing"] + " (" + df["class"] + ")"

        df = df.groupby(["thing", "verb", "success"]).count()

        df.reset_index(inplace=True)

        fig = px.bar(
            df,
            x="verb",
            y="host",
            color="success",
            barmode="stack",
            facet_col="thing",
            facet_col_wrap=facet_col_wrap,
            category_orders={"success": [cat_ok, cat_er]})

        fig.update_layout(title_text="Total interaction counts")
        fig.update_yaxes(title_text="Total count")
        fig.update_xaxes(title_text="Interaction verb")

        return fig

    async def _build_req_latency_fig(self, task, facet_col_wrap, cls_name):
        df = await self._get_thing_df(task=task)

        if df.empty:
            return None

        df.reset_index(inplace=True)

        df = df[df["verb"].isin([
            InteractionVerbs.INVOKE_ACTION,
            InteractionVerbs.WRITE_PROPERTY,
            InteractionVerbs.READ_PROPERTY
        ])]

        df = df[df["class"] == cls_name]

        df["name_ext"] = df["name"] + " (" + df["verb"] + ")"

        if df.empty:
            return None

        fig = px.box(
            df,
            x="name_ext",
            y="latency",
            facet_col="thing",
            facet_col_wrap=facet_col_wrap)

        fig.update_yaxes(title_text="Latency (s)")
        fig.update_xaxes(title_text="Interaction name & verb")

        return fig

    async def build_consumed_request_latency_figure(self, task, facet_col_wrap=2):
        fig = await self._build_req_latency_fig(
            task=task,
            facet_col_wrap=facet_col_wrap,
            cls_name=wotpy.wot.consumed.thing.ConsumedThing.__name__)

        if not fig:
            return None

        fig.update_layout(
            title_text="Latency distribution of consumed properties and actions")

        return fig

    async def build_exposed_request_latency_figure(self, task, facet_col_wrap=2):
        fig = await self._build_req_latency_fig(
            task=task,
            facet_col_wrap=facet_col_wrap,
            cls_name=wotpy.wot.exposed.thing.ExposedThing.__name__)

        if not fig:
            return None

        fig.update_layout(
            title_text="Local handler latency distribution of exposed properties and actions")

        return fig

    async def _build_events_figure(self, task, cls_name, freq, facet_col_wrap, base_height):
        df = await self._get_thing_df(task=task)

        if df.empty:
            return None

        df.reset_index(inplace=True)

        df = df[df["verb"].isin([
            InteractionVerbs.SUBSCRIBE_EVENT
        ])]

        df = df[df["class"] == cls_name]

        if df.empty:
            return None

        event_types = df["event"].unique()

        for event_type in event_types:
            df[event_type] = df["event"] == event_type
            df[event_type].replace(False, np.nan, inplace=True)

        df["name"] = df["name"] + " (" + df["thing"] + ")"

        df.set_index(["date"], inplace=True)

        df = df.groupby([
            pd.Grouper(freq=freq),
            pd.Grouper("name")
        ]).count()

        df = df.filter(event_types)
        df.reset_index(inplace=True)

        num_facets = len(df["name"].unique())

        height = math.ceil(float(num_facets) / facet_col_wrap) * base_height
        height = int(height)

        fig = px.line(
            df,
            x="date",
            y=event_types,
            facet_col="name",
            facet_col_wrap=facet_col_wrap,
            height=height)

        fig.update_yaxes(title_text="Total count")
        fig.update_xaxes(title_text="Date (UTC)")

        return fig

    async def build_consumed_events_figure(self, task, freq="5s", facet_col_wrap=2, base_height=400):
        fig = await self._build_events_figure(
            task=task,
            cls_name=wotpy.wot.consumed.thing.ConsumedThing.__name__,
            freq=freq,
            facet_col_wrap=facet_col_wrap,
            base_height=base_height)

        if not fig:
            return None

        fig.update_layout(
            title_text=f"Timeline of consumed events ({freq} windows)")

        return fig

    async def build_exposed_events_figure(self, task, freq="5s", facet_col_wrap=2, base_height=400):
        fig = await self._build_events_figure(
            task=task,
            cls_name=wotpy.wot.exposed.thing.ExposedThing.__name__,
            freq=freq,
            facet_col_wrap=facet_col_wrap,
            base_height=base_height)

        if not fig:
            return None

        fig.update_layout(
            title_text=f"Timeline of exposed events ({freq} windows)")

        return fig

    async def build_cpu_ranking_figure(self, height_task=45, height_facet=50):
        task_keys = await self._get_tasks()

        if not task_keys or len(task_keys) == 0:
            return None

        info_map = await self._get_info_map()

        dfs = []

        for task in task_keys:
            df = await self._get_system_df(task=task)
            df.reset_index(inplace=True)
            df["task"] = shorten_task_name(task)
            df["task_short"] = shorten_task_name(task)
            task_info = info_map.get(task, {})
            df["cpu_model"] = task_info.get("cpu_model", "Unknown CPU")
            dfs.append(df)

        df_ranking = pd.concat(dfs)

        if df_ranking.empty:
            return None

        models = df_ranking["cpu_model"].unique()

        df_models = [
            df_ranking[df_ranking["cpu_model"] == model]
            for model in models
        ]

        num_tasks = len(df_ranking["task"].unique())

        row_heights = [
            float(len(df_model["task"].unique())) / num_tasks
            for df_model in df_models
        ]

        fig = make_subplots(rows=len(models), cols=1, row_heights=row_heights)

        for idx, model in enumerate(models):
            df_model = df_models[idx]

            trace = go.Box(
                x=df_model["cpu_percent"],
                y=df_model["task_short"],
                name=model)

            fig.add_trace(trace, row=(idx + 1), col=1)

        fig.update_traces(orientation="h")

        height = (len(task_keys) * height_task) + (height_facet * len(models))
        height = max(height, _MIN_HEIGHT)

        fig.update_layout(
            height=height,
            title_text="CPU distribution (by CPU model)",
            showlegend=True)

        fig.update_yaxes(title_text="Task", categoryorder="median ascending")
        fig.update_xaxes(title_text="CPU (%)")

        return fig

    async def build_mem_ranking_figure(self, height_task=45):
        task_keys = await self._get_tasks()

        if not task_keys or len(task_keys) == 0:
            return None

        dfs = []

        for task in task_keys:
            df = await self._get_system_df(task=task)
            df.reset_index(inplace=True)
            df["task"] = shorten_task_name(task)
            df["task_short"] = shorten_task_name(task)
            dfs.append(df)

        df_ranking = pd.concat(dfs)

        if df_ranking.empty:
            return None

        height = height_task * len(task_keys)
        height = max(height, _MIN_HEIGHT)

        fig = px.box(
            df_ranking,
            x="mem_mb",
            y="task_short",
            height=height)

        fig.update_yaxes(title_text="Task", categoryorder="median ascending")
        fig.update_xaxes(title_text="Memory (MB)")
        fig.update_layout(title_text="Memory distribution")

        return fig

    async def build_task_timeline_figure(self, height_task=45):
        df_snap = await self._get_snapshot_df()

        df_snap["finish"] = np.where(
            df_snap["is_running"] == True,
            df_snap["stopped_at"],
            df_snap["updated_at"])

        df_snap["task_short"] = df_snap["task"].apply(shorten_task_name)

        height = height_task * len(df_snap["task"].unique())
        height = max(height, _MIN_HEIGHT)

        fig = px.timeline(
            df_snap,
            x_start="created_at",
            x_end="finish",
            y="task_short",
            color="is_error",
            height=height,
            labels={"is_error": "Error"},
            category_orders={"is_error": [False, True]})

        fig.update_yaxes(title_text="Task")
        fig.update_xaxes(title_text="Date")
        fig.update_layout(title_text="Timeline of task containers")

        return fig

    async def _get_task_section_component(self, task):
        fig_mem = await self.build_task_mem_figure(task=task)
        fig_cpu = await self.build_task_cpu_figure(task=task)
        fig_packet_iface = await self.build_task_packet_iface_figure(task=task)
        fig_packet_proto = await self.build_task_packet_protocol_figure(task=task)
        fig_thing_counts = await self.build_thing_counts_figure(task=task)
        fig_cons_req_lat = await self.build_consumed_request_latency_figure(task=task)
        fig_exps_req_lat = await self.build_exposed_request_latency_figure(task=task)
        fig_cons_events = await self.build_consumed_events_figure(task=task)
        fig_exps_events = await self.build_exposed_events_figure(task=task)

        df_snap = await self._get_snapshot_df()
        snapshot = df_snap[df_snap["task"] == task].to_dict("records")

        if len(snapshot) == 0:
            _logger.warning("Undefined snapshot data for: %s", task)

        if len(snapshot) > 1:
            _logger.warning("Multiple snapshot rows for: %s", task)

        snapshot = snapshot[-1] if len(snapshot) > 0 else None
        info = await self._get_info(task, latest=True)

        return TaskSectionComponent(
            fig_mem=fig_mem,
            fig_cpu=fig_cpu,
            fig_packet_iface=fig_packet_iface,
            fig_packet_proto=fig_packet_proto,
            fig_thing_counts=fig_thing_counts,
            fig_cons_req_lat=fig_cons_req_lat,
            fig_exps_req_lat=fig_exps_req_lat,
            fig_cons_events=fig_cons_events,
            fig_exps_events=fig_exps_events,
            snapshot=snapshot,
            info=info,
            title=task)

    async def _get_service_traffic_component(self):
        fig_inbound = await self.build_service_traffic_figure(inbound=True)
        fig_outbound = await self.build_service_traffic_figure(inbound=False)

        elements = []

        if fig_inbound:
            elements.append(FigureBlockComponent(
                fig_inbound,
                title="Service traffic heatmap (inbound)",
                height=None))

        if fig_outbound:
            elements.append(FigureBlockComponent(
                fig_outbound,
                title="Service traffic heatmap (outbound)",
                height=None))

        return ContainerComponent(elements=elements)

    async def _get_system_ranking_component(self):
        fig_cpu = await self.build_cpu_ranking_figure()
        fig_mem = await self.build_mem_ranking_figure()

        elements = []

        if fig_cpu:
            elements.append(FigureBlockComponent(
                fig_cpu,
                title="CPU usage ranking",
                height=None))

        if fig_mem:
            elements.append(FigureBlockComponent(
                fig_mem,
                title="Memory usage ranking",
                height=None))

        return ContainerComponent(elements=elements)

    async def _get_header_component(self):
        title = lxml.etree.Element("h1", attrib={"class": "display-4"})
        title.text = "WoTemu report"
        sub = lxml.etree.Element("p", attrib={"class": "text-muted lead mb-0"})
        now = datetime.datetime.utcnow()
        sub.text = "Built at UTC {}".format(now.isoformat())

        return ContainerComponent(elements=[title, sub, lxml.etree.Element("hr")])

    async def _get_timeline_component(self):
        fig = await self.build_task_timeline_figure()
        title = "Tasks lifetime"
        elements = [FigureBlockComponent(fig, title=title)] if fig else []
        return ContainerComponent(elements=elements)

    async def _get_compose_component(self):
        compose_dict = await self._reader.get_compose_dict()

        if not compose_dict:
            return None

        return CodeBlockComponent(
            compose_dict,
            title="Compose file of the emulation stack")

    async def _get_network_traffic_component(self):
        fig = await self.build_network_traffic_figure()
        title = "Network traffic"
        elements = [FigureBlockComponent(fig, title=title)] if fig else []
        return ContainerComponent(elements=elements)

    async def build_report(self):
        ini = time.time()

        pages = {}

        tasks = await self._get_tasks()
        tasks = sorted(tasks)

        task_pages = {}

        for task in tasks:
            task_section = await self._get_task_section_component(task=task)
            task_pages[f"{task}.html"] = task_section.to_page_html()

        df_snapshot = await self._get_snapshot_df()
        task_infos = await self._get_info_map()

        task_list = TaskListComponent(
            task_keys=tasks,
            task_infos=task_infos,
            df_snapshot=df_snapshot)

        service_traffic = await self._get_service_traffic_component()
        network_traffic = await self._get_network_traffic_component()
        system_ranking = await self._get_system_ranking_component()
        header = await self._get_header_component()
        timeline = await self._get_timeline_component()
        compose_comp = await self._get_compose_component()

        elements = [header]

        if compose_comp:
            pages.update({"compose.html": compose_comp.to_page_html()})

            compose_link = lxml.etree.Element("a", attrib={
                "class": "btn btn-outline-primary mb-3",
                "href": "compose.html",
                "target": "_blank"
            })

            compose_link.text = "Compose file"

            elements.append(compose_link)

        elements.extend([
            service_traffic,
            network_traffic,
            system_ranking,
            timeline,
            task_list
        ])

        container = ContainerComponent(elements=elements)

        pages.update({"index.html": container.to_page_html()})
        pages.update(task_pages)

        _logger.info("Report built in %s s.", round(time.time() - ini, 2))

        return pages

    async def write_report(self, base_path):
        report_pages = await self.build_report()

        for file_name, file_bytes in report_pages.items():
            file_path = os.path.join(base_path, file_name)

            with open(file_path, "wb") as fh:
                fh.write(file_bytes)

    def _json_df(self, df, orient="records", date_format="iso"):
        if df is None or df.empty:
            return None

        return json.loads(df.to_json(
            orient=orient,
            date_format=date_format))

    async def write_report_dataset(
            self, base_path, file_name=None,
            orient="split", date_format="iso", indent=4):
        task_ids = await self._get_tasks()
        tasks_data = {}

        json_df = functools.partial(
            self._json_df,
            orient=orient,
            date_format=date_format)

        for task_id in task_ids:
            df_system = await self._get_system_df(task=task_id)
            df_packet = await self._get_packet_df(task=task_id, extended=True)
            df_interactions = await self._get_thing_df(task=task_id)
            info = await self._get_info(task_id, latest=True)

            tasks_data[task_id] = {
                "system": json_df(df_system),
                "packet": json_df(df_packet),
                "interaction": json_df(df_interactions),
                "info": info
            }

        df_inb = await self._get_service_traffic_df(inbound=True)
        df_out = await self._get_service_traffic_df(inbound=False)
        df_snap = await self._get_snapshot_df()

        app_metrics = await self._reader.get_app_metrics()

        content = {
            "service_traffic": {
                "inbound": json_df(df_inb),
                "outbound": json_df(df_out)
            },
            "tasks": tasks_data,
            "snapshot": json_df(df_snap),
            "app_metrics": app_metrics
        }

        tstamp = int(time.time() * 1e3)
        file_name = f"{file_name}.json" if file_name else f"wotemu_{tstamp}.json"
        file_path = os.path.join(base_path, file_name)

        with open(file_path, "w") as fh:
            fh.write(json.dumps(content, indent=indent, sort_keys=True))
